"""
The `mod_int` Data Type

* In C/C++ headers, you can `#include "mod_int.h"`
* In Cython files, use `from sage.ext.mod_int cimport *`
"""

# ***************************************************************************
#       Copyright (C) 2013 Volker Braun <vbraun.name@gmail.com>
#       Copyright (C) 2013 William Stein <wstein@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************


cdef extern from "mod_int.h":
    ctypedef long mod_int
    mod_int MOD_INT_MAX
    mod_int MOD_INT_OVERFLOW



r"""
`p`-adic Elements with lattice precision.

AUTHOR:

- Xavier Caruso (2018-02): initial version

"""

# ****************************************************************************
#       Copyright (C) 2018 Xavier Caruso <xavier.caruso@normalesup.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
# ****************************************************************************


from sage.misc.abstract_method import abstract_method

from sage.rings.integer import Integer

from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.rings.infinity import Infinity
from sage.structure.richcmp import rich_to_bool, richcmp
from sage.rings.padics.padic_generic_element import pAdicGenericElement
from sage.rings.padics.lattice_precision import pRational

from sage.rings.padics.precision_error import PrecisionError


def unpickle_le(parent, value, prec):
    r"""
    Unpickle `p`-adic elements.

    INPUT:

    - ``parent`` -- the parent, a `p`-adic ring

    - ``value`` -- a rational number

    - ``prec`` -- integer

    EXAMPLES::

        sage: from sage.rings.padics.padic_lattice_element import unpickle_le
        sage: R = ZpLC(5,8)
        doctest:...: FutureWarning:...
        sage: a = unpickle_le(R, 42, 6); a
        2 + 3*5 + 5^2 + O(5^6)
        sage: a.parent() is R
        True
    """
    return parent(value, prec)


class pAdicLatticeElement(pAdicGenericElement):
    r"""
    Construct new element with given parent and value.

    INPUT:

    - ``parent`` -- the parent of this element

    - ``x`` -- the newly created element

    - ``prec`` -- integer; the absolute precision at which this
      element has to be capped

    - ``dx`` -- dictionary representing the differential of ``x``

    - ``dx_mode`` -- string; either ``'linear_combination'`` (the default)
      or ``'values'``

    - ``valuation`` -- integer or ``None`` (default: ``None``);
      the valuation of this element

    - ``check`` -- boolean (default: ``True``); whether the function
      should check that the given values are well formed and coherent

    - ``reduce`` -- boolean (default: ``True``); whether the given
      values need to be reduced

    TESTS::

        sage: R = ZpLC(2)
        sage: x = R(1, 10)  # indirect doctest
        sage: x
        1 + O(2^10)
    """
    def __init__(self, parent, x, prec=None, dx=[], dx_mode='linear_combination', valuation=None, check=True, reduce=True):
        r"""
        TESTS::

            sage: R = ZpLC(2)
            sage: x = R(1, 10)  # indirect doctest
            sage: x
            1 + O(2^10)
        """
        self._parent = parent
        p = parent.prime()
        pAdicGenericElement.__init__(self, parent)
        self._precision = parent.precision()
        if check:
            if isinstance(x, pAdicGenericElement):
                if parent.prime() != x.parent().prime():
                    raise TypeError("conversion between different p-adic rings/fields not supported")
                if prec is None:
                    prec = x.precision_absolute()
                else:
                    prec = min(prec, x.precision_absolute())
            x = QQ(x)
        if isinstance(x, pRational):
            self._value = x
        else:
            self._value = pRational(p, QQ(x))
        trunc = self._declare_new_element(dx, prec, dx_mode)
        if reduce:
            self._value = self._value.reduce(trunc)

    @abstract_method
    def _declare_new_element(self, dx, prec, dx_mode):
        r"""
        Declare this element to the precision object and
        return the precision at which this element can be truncated safely.

        Only for internal use.

        TESTS::

            sage: R = ZpLC(17)
            sage: prec = R.precision()

            sage: prec.del_elements()
            sage: nb = len(prec.tracked_elements())
            sage: x = R(1, 10)    # indirect doctest
            sage: len(prec.tracked_elements()) == nb + 1
            True
        """
        pass

    def _cache_key(self):
        r"""
        Return a hash of this element.

        TESTS::

            sage: R = ZpLC(2)
            sage: x = R(1, 10)
            sage: x._cache_key()  # random
            140533063823184
        """
        return id(self)

    def __reduce__(self):
        r"""
        Return a tuple of a function and data that can be used to unpickle this
        element.

        EXAMPLES::

            sage: R = ZpLC(5)
            sage: a = R(-3)
            sage: loads(dumps(a)) == a
            True

        For now, diffused digits of precision are not preserved by pickling::

            sage: x, y = R(1, 10), R(1, 5)
            sage: u, v = x+y, x-y
            sage: u + v
            2 + O(5^10)

            sage: up = loads(dumps(u))
            sage: vp = loads(dumps(v))
            sage: up + vp
            2 + O(5^5)
        """
        return unpickle_le, (self.parent(), self.value(), self.precision_absolute())

    def _is_base_elt(self, p):
        r"""
        Return ``True`` if this element is an element of Zp or Qp (rather than
        an extension).

        INPUT:

        - ``p`` -- a prime, which is compared with the parent of this element

        EXAMPLES::

            sage: K = QpLC(7)
            doctest:...: FutureWarning:...
            sage: K.random_element()._is_base_elt(7)  # not tested, known bug (see :issue:`32126`)
            True
        """
        return p == self._parent.prime()

    def approximation(self):
        r"""
        Return an approximation of this element at
        its absolute precision.

        EXAMPLES::

            sage: R = ZpLC(2, print_mode='terse')
            sage: x = R(1234, 10); x
            210 + O(2^10)
            sage: x.approximation()
            210
        """
        prec = self.precision_absolute()
        app = self._value.reduce(prec)
        return app.value()

    def value(self):
        r"""
        Return the actual approximation of this element
        stored in memory.
        In presence of diffused digits of precision, it can
        have more precision than the absolute precision of
        the element.

        EXAMPLES::

            sage: R = ZpLC(2, print_mode='terse')
            sage: x = R(1234, 10); x
            210 + O(2^10)
            sage: x.approximation()
            210

        Another example with diffused digits::

            sage: x = R(2, 10); y = R(7, 5)
            sage: u = x - y
            sage: u
            27 + O(2^5)
            sage: u.value()
            1048571
        """
        return self._value.value()

    def residue(self, absprec=1, field=None, check_prec=True):
        r"""
        Reduces this element modulo `p^{\mathrm{absprec}}`.

        INPUT:

        - ``absprec`` -- nonnegative integer (default: 1)

        - ``field`` -- boolean (default: ``None``); whether to return an
          element of GF(p) or Zmod(p)

        - ``check_prec`` -- boolean (default: ``True``); whether to raise an
          error if this element has insufficient precision to determine the
          reduction

        OUTPUT:

        This element reduced modulo `p^\mathrm{absprec}` as an element of
        `\ZZ/p^\mathrm{absprec}\ZZ`

        EXAMPLES::

            sage: R = ZpLC(7,4)
            sage: a = R(8)
            sage: a.residue(1)
            1

        TESTS::

            sage: R = ZpLC(7,4)
            sage: a = R(8)
            sage: a.residue(0)
            0
            sage: a.residue(-1)
            Traceback (most recent call last):
            ...
            ValueError: cannot reduce modulo a negative power of p
            sage: a.residue(5)
            Traceback (most recent call last):
            ...
            PrecisionError: not enough precision known in order to compute residue
            sage: a.residue(5, check_prec=False)
            8

            sage: a.residue(field=True).parent()
            Finite Field of size 7
        """
        if not isinstance(absprec, Integer):
            absprec = Integer(absprec)
        if check_prec and absprec > self.precision_absolute():
            raise PrecisionError("not enough precision known in order to compute residue")
        elif absprec < 0:
            raise ValueError("cannot reduce modulo a negative power of p")
        if self.valuation() < 0:
            raise ValueError("element must have nonnegative valuation in order to compute residue")
        if field is None:
            field = (absprec == 1)
        elif field and absprec != 1:
            raise ValueError("field keyword may only be set at precision 1")
        p = self._parent.prime()
        if field:
            from sage.rings.finite_rings.finite_field_constructor import GF
            ring = GF(p)
        else:
            from sage.rings.finite_rings.integer_mod_ring import Integers
            ring = Integers(p**absprec)
        return ring(self.value())

    def precision_lattice(self):
        r"""
        Return the precision object (which is a lattice in a possibly
        high-dimensional vector space) that handles the precision of
        this element.

        EXAMPLES::

            sage: R = ZpLC(2, label='precision')
            sage: x = R.random_element()
            sage: y = R.random_element()
            sage: x.precision_lattice()
            Precision lattice on 2 objects (label: precision)

        .. SEEALSO::

            :class:`sage.rings.padics.lattice_precision.PrecisionLattice`
        """
        return self._precision

    def precision_absolute(self):
        r"""
        Return the absolute precision of this element.

        This precision is computed by projecting the lattice precision
        onto the coordinate defined by this element.

        EXAMPLES::

            sage: R = ZpLC(2, print_mode='terse')
            sage: x = R(1234, 10); x
            210 + O(2^10)
            sage: x.precision_absolute()
            10

        Another example with diffused digits::

            sage: x = R(1, 10); y = R(1, 5)
            sage: x, y = x+y, x-y
            sage: x.precision_absolute()
            5
            sage: y.precision_absolute()
            5
            sage: (x+y).precision_absolute()
            11
        """
        prec = self._precision._precision_absolute(self)
        cap = self._value.valuation() + self._parent._prec_cap_relative
        return min(prec, cap)

    def is_precision_capped(self):
        r"""
        Return whether the absolute precision on this element results from a
        cap coming from the parent.

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: x = R(1, 10); x
            1 + O(2^10)
            sage: x.is_precision_capped()
            False

            sage: y = x-x; y
            O(2^40)
            sage: y.is_precision_capped()
            True

            sage: y = x << 35; y
            2^35 + O(2^40)
            sage: y.is_precision_capped()
            True
            sage: z = y >> 35; z
            1 + O(2^5)
            sage: z.is_precision_capped()
            True
        """
        return self._precision._is_precision_capped(self)

    def valuation(self, secure=False):
        r"""
        Return the valuation of this element.

        INPUT:

        - ``secure`` -- boolean (default: ``False``); when ``True``,
          an error is raised if the precision on the element is not
          enough to determine for sure its valuation. Otherwise the
          absolute precision (which is the smallest possible valuation)
          is returned.

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: x = R(12, 10); x
            2^2 + 2^3 + O(2^10)
            sage: x.valuation()
            2

            sage: y = x - x; y
            O(2^40)
            sage: y.valuation()
            40
            sage: y.valuation(secure=True)
            Traceback (most recent call last):
            ...
            PrecisionError: not enough precision
        """
        val = self._value.valuation()
        prec = self.precision_absolute()
        if val < prec:
            return val
        if secure:
            raise PrecisionError("not enough precision")
        else:
            return prec

    def precision_relative(self, secure=False):
        r"""
        Return the relative precision of this element, that is
        the difference between its absolute precision and its
        valuation.

        INPUT:

        - ``secure`` -- boolean (default: ``False``); when ``True``,
          an error is raised if the precision on the element is not
          enough to determine for sure its valuation

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: x = R(12, 10); x
            2^2 + 2^3 + O(2^10)
            sage: x.precision_relative()
            8

            sage: y = x - x; y
            O(2^40)
            sage: y.precision_relative()
            0
            sage: y.precision_relative(secure=True)
            Traceback (most recent call last):
            ...
            PrecisionError: not enough precision
        """
        if not secure and self.is_zero():
            return ZZ(0)
        return self.precision_absolute() - self.valuation(secure=secure)

    def _richcmp_(self, other, op):
        r"""
        Compare this element with ``other``.

        TESTS::

            sage: R = ZpLC(2)
            sage: x = R(1, 5)
            sage: y = R(128, 10)
            sage: z = x + y

            sage: x
            1 + O(2^5)
            sage: z
            1 + O(2^5)

            sage: x == z   # Indirect doctest
            False
            sage: z - x
            2^7 + O(2^10)
        """
        if (self - other).is_zero():
            return rich_to_bool(op, 0)
        return richcmp(QQ(self.lift()), QQ(other.lift()), op)

    def is_equal_to(self, other, prec):
        r"""
        Return ``True`` if this element is indistinguishable
        from ``other`` at precision ``prec``

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: x = R(1, 5)
            sage: y = R(128, 10)
            sage: z = x + y

            sage: x
            1 + O(2^5)
            sage: z
            1 + O(2^5)

            sage: x.is_equal_to(z, 5)
            True

            sage: x.is_equal_to(z, 10)
            False
            sage: z - x
            2^7 + O(2^10)
        """
        return (self-other).is_zero(prec)

    def _add_(self, other):
        r"""
        Return the sum of this element and ``other``.

        EXAMPLES::

            sage: R = ZpLC(19, 5)
            sage: a = R(-1); a
            18 + 18*19 + 18*19^2 + 18*19^3 + 18*19^4 + O(19^5)
            sage: b = R(-5/2); b
            7 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)
            sage: a + b   # indirect doctest
            6 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)

        TESTS::

            sage: a = R.random_element()
            sage: b = R.random_element()
            sage: a + b == b + a
            True
        """
        x = self._value + other._value
        # elements whose valuation are not less than _zero_cap are assumed to vanish
        # (_zero_cap is set at the creation of the parent)
        if self._parent._zero_cap is not None:
            if x.valuation() >= min(self._value.valuation(), other._value.valuation()) + self._parent._zero_cap:
                x = self._parent._approx_zero
        dx = [  [self, self._parent._approx_one],
               [other, self._parent._approx_one] ]
        return self.__class__(self._parent, x, dx=dx, check=False)

    def _sub_(self, other):
        r"""
        Return the difference of this element and ``other``.

        EXAMPLES::

           sage: R = ZpLC(19, 5)
           sage: a = R(-1); a
           18 + 18*19 + 18*19^2 + 18*19^3 + 18*19^4 + O(19^5)
           sage: b = R(-5/2); b
           7 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)
           sage: a - b   # indirect doctest
           11 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)
        """
        x = self._value - other._value
        if self._parent._zero_cap is not None:
            if x.valuation() >= min(self._value.valuation(), other._value.valuation()) + self._parent._zero_cap:
                x = self._parent._approx_zero
        dx = [  [self, self._parent._approx_one],
               [other, self._parent._approx_minusone] ]
        return self.__class__(self._parent, x, dx=dx, check=False)

    def _mul_(self, other):
        r"""
        Return the product of this element and ``other``.

        EXAMPLES::

            sage: R = ZpLC(19, 5)
            sage: a = R(-1); a
            18 + 18*19 + 18*19^2 + 18*19^3 + 18*19^4 + O(19^5)
            sage: b = R(-5/2); b
            7 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)
            sage: a * b   # indirect doctest
            12 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)

        TESTS::

            sage: a = R.random_element()
            sage: b = R.random_element()
            sage: a * b == b * a
            True

            sage: a = R.random_element()
            sage: b = R.random_element()
            sage: c = R.random_element()
            sage: a * (b+c) == a*b + a*c
            True
        """
        x_self = self._value
        x_other = other._value
        x = x_self * x_other
        dx = [  [self, x_other],
               [other, x_self ] ]
        return self.__class__(self._parent, x, dx=dx, check=False)

    def _div_(self, other):
        r"""
        Return the quotient of this element and ``other``.

        .. NOTE::

            The result of division always lives in the fraction field,
            even if the element to be inverted is a unit.

        EXAMPLES::

            sage: R = ZpLC(19)
            sage: a = R(-1, 5); a
            18 + 18*19 + 18*19^2 + 18*19^3 + 18*19^4 + O(19^5)
            sage: b = R(-5/2, 5); b
            7 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)

            sage: c = a / b   # indirect doctest
            sage: c
            8 + 11*19 + 7*19^2 + 11*19^3 + 7*19^4 + O(19^5)
            sage: c.parent()
            19-adic Field with lattice-cap precision

            sage: a / (19*b)
            8*19^-1 + 11 + 7*19 + 11*19^2 + 7*19^3 + O(19^4)

        If the division is indistinguishable from zero, an error is raised::

            sage: c = a / (2*b + 5)
            Traceback (most recent call last):
            ...
            PrecisionError: cannot divide by something indistinguishable from zero
        """
        if other.is_zero():
            raise PrecisionError("cannot divide by something indistinguishable from zero")
        x_self = self._value
        x_other = other._value
        x = x_self / x_other
        # dx = (1/other)*dself - (self/other^2)*dother
        dx = [  [self, self._parent._approx_one/x_other],
               [other, -x_self/(x_other*x_other)] ]
        return self.__class__(self._parent.fraction_field(), x, dx=dx, check=False)

    def __invert__(self):
        r"""
        Return the multiplicative inverse of this element.

        .. NOTE::

            The result of division always lives in the fraction field,
            even if the element to be inverted is a unit.

        EXAMPLES::

            sage: R = ZpLC(19)
            sage: x = R(-5/2, 5); x
            7 + 9*19 + 9*19^2 + 9*19^3 + 9*19^4 + O(19^5)

            sage: y = ~x    # indirect doctest
            sage: y
            11 + 7*19 + 11*19^2 + 7*19^3 + 11*19^4 + O(19^5)
            sage: y == -2/5
            True

        TESTS::

            sage: a = R.random_element()
            sage: a * ~a == 1
            True
        """
        if self.is_zero():
            raise PrecisionError("cannot invert something indistinguishable from zero")
        x_self = self._value
        x = self._parent._approx_one / x_self
        # dx = -(1/self^2)*dself
        dx = [  [self, self._parent._approx_minusone/(x_self*x_self)] ]
        return self.__class__(self._parent.fraction_field(), x, dx=dx, check=False)

    def _quo_rem(self, other):
        """
        Quotient with remainder.

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: a = R(373286)
            sage: b = R(12685856)
            sage: q,r = a.quo_rem(b); q, r
            (1 + 2^8 + 2^12 + 2^15 + O(2^16), 2 + 2^2 + O(2^21))
            sage: q*b+r == a
            True
            sage: q,r = b.quo_rem(a); q, r
            (2^4 + 2^5 + 2^7 + 2^10 + 2^14 + 2^15 + 2^16 + 2^17 + O(2^24), O(2^40))
            sage: q*a == b
            True
        """
        if other.is_zero():
            # We use ZeroDivisionError since _test_quo_rem expects it.
            raise ZeroDivisionError("cannot divide by something indistinguishable from zero")
        if other.valuation() > self.precision_absolute():
            raise PrecisionError
        q, r = self._value._quo_rem(other._value)
        rem = self.__class__(self._parent, r, check=False)
        quo = self.parent()((self - rem) / other)
        return quo, rem

    def add_bigoh(self, prec):
        r"""
        Return a new element with absolute precision decreased to
        the specified precision.

        INPUT:

        - ``prec`` -- integer or infinity

        EXAMPLES::

           sage: R = ZpLC(7)
           sage: a = R(8); a.add_bigoh(1)
           1 + O(7)
           sage: b = R(0); b.add_bigoh(3)
           O(7^3)

           sage: R = QpLC(7, 4)
           sage: a = R(8); a.add_bigoh(1)
           1 + O(7)
           sage: b = R(0); b.add_bigoh(3)
           O(7^3)

           The precision never increases::

           sage: R(4).add_bigoh(2).add_bigoh(4)
           4 + O(7^2)

        If ``prec`` is negative, the output is an element of the
        fraction field::

           sage: c = a.add_bigoh(-1); c
           O(7^-1)
           sage: c.parent()
           7-adic Field with lattice-cap precision
        """
        if prec is Infinity:
            return self
        if not self._parent.is_field() and prec < 0:
            field = self._parent.fraction_field()
            return self._copy(field).add_bigoh(prec)
        x = self._value
        dx = [ [self, self._parent._approx_one ] ]
        return self.__class__(self._parent, x, prec, dx=dx, check=False)

    def lift_to_precision(self, prec=None, infer_precision=False):
        r"""
        Return another element of the same parent with absolute precision
        at least ``prec``, congruent to this `p`-adic element modulo the
        precision of this element.

        INPUT:

        - ``prec`` -- integer or ``None`` (default: ``None``); the
          absolute precision of the result. If ``None``, lifts to the
          maximum precision allowed.

        - ``infer_precision`` -- boolean (default: ``False``)

        NOTE:

        In the lattice precision model, the precision of all variables is
        handled globally by a unique object, namely a lattice in a certain
        vector space.

        When ``infer_precision`` is set to ``True``, the precision lattice
        is recomputed. This may affect the precision of other variables
        with the same parent.

        When ``infer_precision`` is set to ``False``, the precision on the
        newly created variable is independent as if the variable were created
        by hand by setting independently the value of the absolute precision.
        In particular, if ``self`` used to share diffused digits of precision
        with other variables, they are not preserved.

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: x = R(1, 10); x
            1 + O(2^10)
            sage: x.lift_to_precision(15)
            1 + O(2^15)
            sage: x.lift_to_precision()
            1 + O(2^20)

        An example with diffused digits of precision::

            sage: x = R(1, 10); y = R(1, 5)
            sage: u = x+y; u
            2 + O(2^5)
            sage: v = x-y; v
            O(2^5)
            sage: u + v
            2 + O(2^11)

        The gain of precision on ``u + v`` is due to the presence of diffused
        digits of precision between ``u`` and ``v``.

        However, if we call :meth:`lift_to_precision` on one of these variables,
        these diffused digits are lost and the precision on the sum is no longer
        sharp::

            sage: u.lift_to_precision() + v
            2 + O(2^5)

        We can avoid this issue as follows::

            sage: u.lift_to_precision(infer_precision=True) + v
            2 + O(2^11)

        But now the precision on ``y`` has changed::

            sage: y
            1 + O(2^10)

        Indeed if the absolute precision on ``u = x+y`` (resp. on ``x``)
        is 20 (resp. 10), we deduce that the absolution precision on
        ``y = u-x`` is 10.

        .. SEEALSO::

            :meth:`lift_to_precision` of the precision object
        """
        #from warnings import warn
        #warn("use lift_to_precision with extreme caution in the framework of lattice precision")
        parent = self._parent
        if infer_precision:
            cap = min(parent.precision_cap_absolute(), parent.precision_cap_relative() + self._value.valuation())
            if prec is None or prec > cap:
                prec = cap
            lift = self._copy()
            parent.precision()._lift_to_precision(lift, prec)
        else:
            lift = self.__class__(parent, self._value, prec, check=False)
        return lift

    def _is_inexact_zero(self):
        r"""
        Return ``True`` if this element is indistinguishable from zero.

        EXAMPLES::

            sage: R = ZpLC(5)
            sage: R(0)._is_inexact_zero()
            True
            sage: R(1)._is_inexact_zero()
            False
        """
        absprec = self.precision_absolute()
        return self._value.valuation() >= absprec

    def is_zero(self, prec=None):
        r"""
        Return ``True`` if this element is indistinguishable from zero
        at the given precision (if given).

        INPUT:

        - ``prec`` -- integer or ``None`` (default: ``None``)

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: x = R(2/5, 10); x
            2 + 2^3 + 2^4 + 2^7 + 2^8 + O(2^10)
            sage: x.is_zero()
            False
            sage: x.is_zero(1)
            True

            sage: (5*x-2).is_zero()
            True
            sage: 5*x == 2   # indirect doctest
            True
        """
        absprec = self.precision_absolute()
        if prec is None:
            prec = absprec
        else:
            prec = min(absprec, prec)
        return self._value.valuation() >= prec

    def lift(self):
        r"""
        Return an integer or rational congruent to this element modulo
        its absolute precision.
        If a rational is returned, its denominator will be a power of `p`.

        EXAMPLES::

           sage: R = ZpLC(7)
           sage: a = R(8); a.lift()
           8

           sage: R = QpLC(7)
           sage: a = R(8); a.lift()
           8
           sage: b = R(8/7); b.lift()
           8/7
        """
        return self._value.value()

    def __rshift__(self, n):
        r"""
        Divide this element by ``p^n``, and truncate
        (if the parent is not a field).

        EXAMPLES::

            sage: R = ZpLC(997, 7)
            sage: a = R(123456878908); a
            964*997 + 572*997^2 + 124*997^3 + O(997^8)

            sage: S = ZpLC(5)
            sage: b = S(17); b
            2 + 3*5 + O(5^20)

        Shifting to the right divides by a power of `p`, but drops
        terms with negative valuation::

            sage: a >> 3
            124 + O(997^5)
            sage: b >> 1
            3 + O(5^19)
            sage: b >> 40
            O(5^0)

        If the parent is a field no truncation is performed::

            sage: K = QpLC(5)
            sage: b = K(17); b
            2 + 3*5 + O(5^20)
            sage: b >> 1
            2*5^-1 + 3 + O(5^19)

        A negative shift multiplies by that power of `p`::

            sage: a >> -3
            964*997^4 + 572*997^5 + 124*997^6 + O(997^11)
            sage: b >> -5
            2*5^5 + 3*5^6 + O(5^25)
        """
        return self << (-n)

    def __lshift__(self, n):
        r"""
        Multiply this element by ``p^n``.

        If ``n`` is negative and this element does not lie in a field,
        digits may be truncated.  See :meth:`__rshift__` for details.

        EXAMPLES::

            sage: R = ZpLC(5)
            sage: a = R(1000); a
            3*5^3 + 5^4 + O(5^23)
            sage: a >> 1
            3*5^2 + 5^3 + O(5^22)

            sage: S = Zp(5); b = S(1000); b
            3*5^3 + 5^4 + O(5^23)
        """
        from sage.rings.padics.generic_nodes import pAdicRingBaseGeneric
        parent = self._parent
        p = parent.prime()
        if isinstance(parent, pAdicRingBaseGeneric):
            if self.precision_absolute() + n < 0:
                return self.__class__(parent, pRational(p, 0), 0, dx={}, check=False)
        powp = pRational(p, ZZ(1), n)
        x = self._value * powp
        if isinstance(parent, pAdicRingBaseGeneric):
            x -= x.reduce(0)
        dx = [ [self, powp] ]
        return self.__class__(parent, x, dx=dx, check=False)

    def unit_part(self):
        r"""
        Return `u`, where this element is `p^v u` and `u` is a unit.

        EXAMPLES::

            sage: R = ZpLC(17)
            sage: a = R(18*17, 4)
            sage: a.unit_part()
            1 + 17 + O(17^3)

            sage: b=1/a; b
            17^-1 + 16 + O(17^2)
            sage: b.unit_part()
            1 + 16*17 + O(17^3)

        If the element is indistinguishable from zero, an error is raised::

            sage: c = R(0, 5); c
            O(17^5)
            sage: c.unit_part()
            Traceback (most recent call last):
            ...
            PrecisionError: not enough precision
        """
        v = self.valuation(secure=True)
        return self >> v

    def val_unit(self):
        r"""
        Return the pair `(v, u)`, where this element is
        `p^v u` and `u` is a unit.

        EXAMPLES::

            sage: R = ZpLC(17)
            sage: a = R(18*17, 4)
            sage: a.val_unit()
            (1, 1 + 17 + O(17^3))

            sage: b=1/a; b
            17^-1 + 16 + O(17^2)
            sage: b.val_unit()
            (-1, 1 + 16*17 + O(17^3))

        If the element is indistinguishable from zero, an error is raised

            sage: c = R(0, 5); c
            O(17^5)
            sage: c.val_unit()
            Traceback (most recent call last):
            ...
            PrecisionError: not enough precision
        """
        v = self.valuation(secure=True)
        return v, self >> v

    def _copy(self, parent=None):
        r"""
        Return a copy of this element or convert this element
        to the given parent provided that the precision on this
        parent is handled by the same precision object.

        EXAMPLES::

            sage: R = ZpLC(2)
            sage: x = R(1, 10); x
            1 + O(2^10)
            sage: y = x._copy()
            sage: y
            1 + O(2^10)

        In the lattice precision model, Sage remembers that ``y`` is
        actually equal to ``x``. Therefore, when we compute the difference,
        the `O(2^10)` cancel as well::

            sage: x - y
            O(2^20)

        This function can also be used for coercion/conversion as follows::

            sage: K = QpLC(2)
            sage: y = x._copy(K)
            sage: y
            1 + O(2^10)
            sage: y.parent()
            2-adic Field with lattice-cap precision

            sage: a = K(2, 10); a
            2 + O(2^10)
            sage: b = a._copy(R)
            sage: b
            2 + O(2^10)
            sage: b.parent()
            2-adic Ring with lattice-cap precision

        In any case, precision is sharp::

            sage: x - y
            O(2^20)
            sage: a - b
            O(2^21)

        If a parent is given, it must share the same precision object::

            sage: x._copy(ZpLC(5))
            Traceback (most recent call last):
            ...
            TypeError: parent must share the same precision object

            sage: x._copy(Zp(2))
            Traceback (most recent call last):
            ...
            TypeError: parent must share the same precision object

            sage: x._copy(ZpLC(2, label='other'))
            Traceback (most recent call last):
            ...
            TypeError: parent must share the same precision object

        TESTS::

            sage: K(1/2)._copy(R)
            Traceback (most recent call last):
            ...
            ValueError: element of negative valuation cannot be converted to the integer ring
        """
        if parent is None:
            parent = self._parent
        else:
            try:
                if parent.precision() is not self._parent.precision():
                    raise TypeError("parent must share the same precision object")
            except AttributeError:
                raise TypeError("parent must share the same precision object")
            from sage.rings.padics.generic_nodes import pAdicRingBaseGeneric
            if isinstance(parent, pAdicRingBaseGeneric) and self.valuation() < 0:
                raise ValueError("element of negative valuation cannot be converted to the integer ring")
        dx = [ [ self, self._parent._approx_one ] ]
        return self.__class__(parent, self._value, dx=dx, check=False)

    def __copy__(self):
        r"""
        Return a copy of this element.

        TESTS::

            sage: R = ZpLC(2)
            sage: x = R(1, 10); x
            1 + O(2^10)
            sage: y = copy(x)   # indirect doctest
            sage: y
            1 + O(2^10)

            sage: x - y
            O(2^20)
        """
        return self._copy()

    def expansion(self, n=None, lift_mode='simple', start_val=None):
        r"""
        Return a list giving the `p`-adic expansion of this element.
        If this is a field element, start at
        `p^{\mbox{valuation}}`, if a ring element at `p^0`.

        INPUT:

        - ``n`` -- integer or ``None`` (default: ``None``); if given,
          return the corresponding entry in the expansion

        - ``lift_mode`` -- string (default: ``'simple'``); currently
          only ``'simple'`` is implemented

        - ``start_val`` -- integer or ``None`` (default: ``None``);
          start at this valuation rather than the default (`0` or the
          valuation of this element).

        EXAMPLES::

            sage: R = ZpLC(5, 10)
            sage: x = R(123456789); x
            4 + 2*5 + 5^2 + 4*5^3 + 5^5 + 5^6 + 5^8 + 3*5^9 + O(5^10)
            sage: x.expansion()
            [4, 2, 1, 4, 0, 1, 1, 0, 1, 3]

            sage: x.expansion(3)
            4

            sage: x.expansion(start_val=5)
            [1, 1, 0, 1, 3]

        If any, trailing zeros are included in the expansion::

            sage: y = R(1234); y
            4 + 5 + 4*5^2 + 4*5^3 + 5^4 + O(5^10)
            sage: y.expansion()
            [4, 1, 4, 4, 1, 0, 0, 0, 0, 0]
        """
        if lift_mode != 'simple':
            raise NotImplementedError("other modes than 'simple' are not implemented yet")
        prec = self.precision_absolute()
        val = self.valuation()
        expansion = self._value.list(prec)
        if n is not None:
            if n < val:
                return ZZ(0)
            try:
                return expansion[n-val]
            except KeyError:
                raise PrecisionError("the digit in position %s is not determined" % n)
        if start_val is None:
            if self._parent.is_field():
                start_val = val
            else:
                start_val = 0
        if start_val > val:
            return expansion[start_val-val:]
        return (val-start_val)*[ZZ(0)] + expansion

    def dist(self, other):
        r"""
        Return the distance between this element and ``other``.
        The distance is normalized so that `dist(0,p) = 1/p`.

        EXAMPLES::

            sage: R = ZpLC(3)
            sage: x = R(1, 5)
            sage: y = R(4, 5)
            sage: x.dist(y)
            1/3

        TESTS::

            sage: z = R(3^7,10)
            sage: x
            1 + O(3^5)
            sage: x + z
            1 + O(3^5)
            sage: x.dist(x+z)
            1/2187
        """
        x = self - other
        p = self._parent.prime()
        if x.is_zero():
            return ZZ(0)
        return p**(-x.valuation())


class pAdicLatticeCapElement(pAdicLatticeElement):
    def _declare_new_element(self, dx, prec, dx_mode):
        r"""
        Declare this element to the precision object and
        return the precision at which this element can be truncated safely.

        Only for internal use.

        TESTS::

            sage: R = ZpLC(17)
            sage: prec = R.precision()

            sage: prec.del_elements()
            sage: nb = len(prec.tracked_elements())
            sage: x = R(1, 10)    # indirect doctest
            sage: len(prec.tracked_elements()) == nb + 1
            True
        """
        parent = self._parent
        cap = min(parent.precision_cap_absolute(), parent.precision_cap_relative() + self._value.valuation())
        if prec is None or prec > cap:
            capped = True
            prec = cap
        else:
            capped = False
        self._precision._new_element(self, dx, bigoh=prec, dx_mode=dx_mode, capped=capped)
        return prec

    def _is_exact_zero(self):
        r"""
        Return ``True`` if this element is exactly zero.

        .. NOTE::

            Since exact zeros are not supported in the precision lattice
            model, this function always returns ``False``.

        EXAMPLES::

            sage: R = ZpLC(5)
            sage: R(0)._is_exact_zero()
            False
            sage: R(1)._is_exact_zero()
            False
        """
        return False


class pAdicLatticeFloatElement(pAdicLatticeElement):
    def _declare_new_element(self, dx, prec, dx_mode):
        r"""
        Declare this element to the precision object and
        return the precision at which this element can be truncated safely.

        Only for internal use.

        TESTS::

            sage: R = ZpLF(17)
            doctest:...: FutureWarning:...
            sage: prec = R.precision()

            sage: prec.del_elements()
            sage: nb = len(prec.tracked_elements())
            sage: x = R(1, 10)    # indirect doctest
            sage: len(prec.tracked_elements()) == nb + 1
            True
        """
        self._precision._new_element(self, dx, bigoh=prec, dx_mode=dx_mode)
        cap = self._precision.internal_prec() + self._value.valuation()
        if prec is None:
            return cap
        return min(cap, prec)

    def _is_exact_zero(self):
        r"""
        Return ``True`` if this element is exactly zero.

        EXAMPLES::

            sage: R = ZpLF(5)
            sage: R(0)._is_exact_zero()
            True
            sage: R(0, 10)._is_exact_zero()
            False

            sage: R(1)._is_exact_zero()
            False
        """
        return self._value.is_zero() and self._precision._precision_absolute(self) is Infinity



# sage.doctest: needs sage.combinat
r"""
Finitely generated free graded left modules over connected graded algebras

Let `A` be a connected graded algebra. Some methods here require in
addition that `A` be an algebra over a field or a PID and that Sage
has a description of a basis for `A`.

For example, let `p` be a prime number. The mod `p` Steenrod algebra
`A_p` is a connected algebra over the finite field of `p` elements.
Many of the modules presented here will be defined over `A_p`, or one
of its sub-Hopf algebras.  E.g.::

    sage: A = SteenrodAlgebra(p=2)

However, the current implementation can use any connected graded
algebra that has a graded basis where each graded part is finite
dimensional. Another good family is the exterior algebras::

    sage: E.<x,y,z> = ExteriorAlgebra(QQ)

A free module is defined by the graded algebra and an ordered tuple
of degrees for the generators::

    sage: M = A.free_graded_module(generator_degrees=(0,1))
    sage: M
    Free graded left module on 2 generators over
     mod 2 Steenrod algebra, milnor basis

    sage: F.<a,b,c> = E.free_graded_module((0,3,6))
    sage: F
    Free graded left module on 3 generators over
     The exterior algebra of rank 3 over Rational Field

The resulting free modules will have generators in the degrees as specified::

    sage: M.generator_degrees()
    (0, 1)
    sage: F.generator_degrees()
    (0, 3, 6)

The default names for the generators are ``g[degree]`` if they are in
distinct degrees, ``g[degree, i]`` otherwise. They can be given other
names, as was done when creating the module ``F``::

    sage: M.generators()
    (g[0], g[1])
    sage: F.generators()
    (a, b, c)

The connectivity of a module over a connected graded algebra is the minimum
degree of all its module generators.  Thus, if the module is non-trivial, the
connectivity is an integer::

    sage: M.connectivity()
    0

Module elements
---------------

For an `A`-module with generators `\{g_i\}_{i=1}^N`, any homogeneous element
of degree `n` has the form

.. MATH::

    x = \sum_{i=1}^N a_i\cdot g_i\,,

where `a_i\in A_{n-\deg(g_i)}` for all `i`.  The ordered set `\{a_i\}`
is referred to as the coefficients of `x`.

You can produce module elements from a given set of coefficients::

    sage: coeffs = [Sq(5), Sq(1,1)]
    sage: x = M(coeffs); x
    Sq(5)*g[0] + Sq(1,1)*g[1]

You can also use the module action::

    sage: Sq(2) * x
    (Sq(4,1)+Sq(7))*g[0] + Sq(3,1)*g[1]

Each nonzero element has a well-defined degree::

    sage: x.degree()
    5

However the zero element does not::

    sage: zero = M.zero(); zero
    0
    sage: zero.degree()
    Traceback (most recent call last):
    ...
    ValueError: the zero element does not have a well-defined degree

Any two elements can be added as long as they are in the same degree::

    sage: y = M.an_element(5); y
    Sq(2,1)*g[0] + Sq(4)*g[1]
    sage: x + y
    (Sq(2,1)+Sq(5))*g[0] + (Sq(1,1)+Sq(4))*g[1]

or when at least one of them is zero::

    sage: x + zero == x
    True

Finally, additive inverses exist::

    sage: x - x
    0

For every integer `n`, the set of module elements of degree `n` form
a free module over the ground ring `k`.  A basis for this free module
can be computed::

    sage: M.basis_elements(5)
    (Sq(2,1)*g[0], Sq(5)*g[0], Sq(1,1)*g[1], Sq(4)*g[1])

together with a corresponding free module presentation::

    sage: M.vector_presentation(5)
    Vector space of dimension 4 over Finite Field of size 2

Given any element, its coordinates with respect to this basis can be computed::

    sage: v = x.vector_presentation(); v
    (0, 1, 1, 0)

Going the other way, any element can be constructed by specifying its
coordinates::

    sage: x_ = M.element_from_coordinates((0,1,1,0), 5)
    sage: x_
    Sq(5)*g[0] + Sq(1,1)*g[1]
    sage: x_ == x
    True


Module homomorphisms
--------------------

Homomorphisms of free graded `A`-modules `M\to N` are linear maps of their
underlying free `k`-module which commute with the `A`-module structure.

To create a homomorphism, first create the object modeling the set of all
such homomorphisms using the free function ``Hom``::

    sage: M = A.free_graded_module((0,1))
    sage: N.<c2> = A.free_graded_module((2,))
    sage: homspace = Hom(M, N); homspace
    Set of Morphisms from Free graded left module on 2 generators
      over mod 2 Steenrod algebra, milnor basis
     to Free graded left module on 1 generator
      over mod 2 Steenrod algebra, milnor basis
     in Category of finite dimensional graded modules with basis
      over mod 2 Steenrod algebra, milnor basis

Just as module elements, homomorphisms are created using the homspace
object. The only argument is a list of module elements in the codomain,
corresponding to the module generators of the domain::

    sage: values = [Sq(2)*c2, Sq(2)*Sq(1)*c2]
    sage: f = homspace(values)

The resulting homomorphism is the one sending the `i`-th generator of the
domain to the `i`-th codomain value given::

    sage: f
    Module morphism:
      From: Free graded left module on 2 generators over mod 2 Steenrod algebra, milnor basis
      To:   Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
      Defn: g[0] |--> Sq(2)*c2
            g[1] |--> (Sq(0,1)+Sq(3))*c2

Convenience methods exist for creating the trivial morphism::

    sage: homspace.zero()
    Module morphism:
      From: Free graded left module on 2 generators over mod 2 Steenrod algebra, milnor basis
      To:   Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
      Defn: g[0] |--> 0
            g[1] |--> 0

as well as the identity endomorphism::

    sage: Hom(M, M).identity()
    Module endomorphism of Free graded left module on 2 generators over mod 2 Steenrod algebra, milnor basis
      Defn: g[0] |--> g[0]
            g[1] |--> g[1]

Homomorphisms can be evaluated on elements of the domain module::

    sage: v1 = f(Sq(7)*M.generator(0)); v1
    Sq(3,2)*c2

    sage: v2 = f(Sq(17)*M.generator(1)); v2
    (Sq(11,3)+Sq(13,0,1)+Sq(17,1))*c2

and they respect the module action::

    sage: v1 == Sq(7)*f(M.generator(0))
    True

    sage: v2 == Sq(17)*f(M.generator(1))
    True

Any non-trivial homomorphism has a well-defined degree::

    sage: f.degree()
    4

but just as module elements, the trivial homomorphism does not::

    sage: zero_map = homspace.zero()
    sage: zero_map.degree()
    Traceback (most recent call last):
    ...
    ValueError: the zero morphism does not have a well-defined degree

Any two homomorphisms can be added as long as they are of the same degree::

    sage: f2 = homspace([Sq(2)*c2, Sq(3)*c2])
    sage: f + f2
    Module morphism:
      From: Free graded left module on 2 generators over mod 2 Steenrod algebra, milnor basis
      To:   Free graded left module on 1 generator over mod 2 Steenrod algebra, milnor basis
      Defn: g[0] |--> 0
            g[1] |--> Sq(0,1)*c2

or when at least one of them is zero::

    sage: f + zero_map == f
    True

Finally, additive inverses exist::

    sage: f - f == 0
    True

The restriction of a homomorphism to the free module of `n`-dimensional
module elements is a linear transformation::

    sage: f_4 = f.vector_presentation(4); f_4
    Vector space morphism represented by the matrix:
    [0 1 0]
    [1 1 1]
    [0 1 0]
    [0 0 0]
    Domain: Vector space of dimension 4 over Finite Field of size 2
    Codomain: Vector space of dimension 3 over Finite Field of size 2

This is compatible with the vector presentations of its domain and codomain
modules::

    sage: f.domain() is M
    True
    sage: f.codomain() is N
    True
    sage: f_4.domain() is M.vector_presentation(4)
    True
    sage: f_4.codomain() is N.vector_presentation(4 + f.degree())
    True


AUTHORS:

- Robert R. Bruner, Michael J. Catanzaro (2012): Initial version.
- Sverre Lunoee--Nielsen and Koen van Woerden (2019-11-29): Updated the
  original software to Sage version 8.9.
- Sverre Lunoee--Nielsen (2020-07-01): Refactored the code and added
  new documentation and tests.
"""

# *****************************************************************************
#       Copyright (C) 2019 Robert R. Bruner <rrb@math.wayne.edu>
#                     and  Michael J. Catanzaro <mike@math.wayne.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.modules.free_module import FreeModule
from sage.modules.fp_graded.free_element import FreeGradedModuleElement
from sage.rings.infinity import infinity
from sage.categories.graded_modules import GradedModules
from sage.categories.principal_ideal_domains import PrincipalIdealDomains
from sage.categories.homset import Hom
from sage.combinat.free_module import CombinatorialFreeModule


class FreeGradedModule(CombinatorialFreeModule):
    r"""
    Create a finitely generated free graded module over a connected
    graded algebra, with generators in specified degrees.

    INPUT:

    - ``algebra`` -- the graded connected algebra over which the module is
      defined; this algebra must be equipped with a graded basis

    - ``generator_degrees`` -- tuple of integers defining the number
      of generators of the module, and their degrees

    - ``names`` -- (optional) the names of the generators. If ``names``
      is a comma-separated string like ``'a, b, c'``, then those will
      be the names. Otherwise, for example if ``names`` is ``abc``,
      then the names will be ``abc(d,i)``.

    By default, if all generators are in distinct degrees, then the
    ``names`` of the generators will have the form ``g_{d}`` where
    ``d`` is the degree of the generator. If the degrees are not
    distinct, then the generators will be called ``g_{d,i}`` where
    ``d`` is the degree and ``i`` is its index in the list of
    generators in that degree.

    EXAMPLES::

        sage: from sage.modules.fp_graded.free_module import FreeGradedModule
        sage: E.<x,y,z> = ExteriorAlgebra(QQ)
        sage: M = FreeGradedModule(E, (-1,3))
        sage: M
        Free graded left module on 2 generators over
         The exterior algebra of rank 3 over Rational Field
        sage: M.generator_degrees()
        (-1, 3)
        sage: a, b = M.generators()
        sage: (x*y*b).degree()
        5

    ``names`` of generators::

        sage: M.generators()
        (g[-1], g[3])
        sage: FreeGradedModule(E, (0, 0, 2)).generators()
        (g[0, 0], g[0, 1], g[2, 0])
        sage: FreeGradedModule(E, (0, 0, 2), names='x, y, z').generators()
        (x, y, z)
        sage: FreeGradedModule(E, (0, 0, 2), names='xyz').generators()
        (xyz[0, 0], xyz[0, 1], xyz[2, 0])

    ``names`` can also be defined implicitly using Sage's ``M.<...>`` syntax::

        sage: A = SteenrodAlgebra(2)
        sage: M.<x,y,z> = FreeGradedModule(A, (-2,2,4))
        sage: M
        Free graded left module on 3 generators over
         mod 2 Steenrod algebra, milnor basis
        sage: M.gens()
        (x, y, z)
    """
    def __classcall__(cls, algebra, generator_degrees, category=None,
                      names=None, prefix=None, **kwds):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: E.<x,y> = ExteriorAlgebra(QQ)
            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: M1 = FreeGradedModule(E, [1, 0, 2], names='a,b,c')
            sage: M2.<a,b,c> = FreeGradedModule(E, (1, 0, 2))
            sage: M1 is M2
            True
        """
        if algebra.base_ring() not in PrincipalIdealDomains():
            raise ValueError('the ground ring of the algebra must be a PID')

        generator_degrees = tuple(generator_degrees)
        category = GradedModules(algebra).WithBasis().FiniteDimensional().or_subcategory(category)
        if names is not None:
            from sage.structure.category_object import normalize_names
            names = normalize_names(-1, names)
            if len(generator_degrees) > 1:
                if len(names) == 1:
                    if prefix is None:
                        prefix = names[0]
                    names = None  # if prefix is specified and takes priority
            if names is not None and len(names) != len(generator_degrees):
                raise ValueError("the names do not correspond to the generators")
        if prefix is None:
            prefix = 'g'
        return super().__classcall__(cls, algebra=algebra,
                                     generator_degrees=generator_degrees,
                                     category=category, names=names,
                                     prefix=prefix, **kwds)

    def __init__(self, algebra, generator_degrees, category, names=None, **kwds):
        r"""
        Create a finitely generated free graded module over a connected graded
        algebra.

        TESTS::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: TestSuite(FreeGradedModule(A, (-2,2,4))).run()
        """
        # If generator_degrees is [d_0, d_1, ...], then
        # the generators are indexed by (0,d_0), (1,d_1), ...
        keys = list(enumerate(generator_degrees))
        self._generator_degrees = generator_degrees

        keys = []
        degs_so_far = {}
        unique = True
        for i in generator_degrees:
            if i in degs_so_far:
                idx = degs_so_far[i] + 1
                degs_so_far[i] += 1
                unique = False
            else:
                idx = 0
                degs_so_far[i] = 0
            keys.append((i, idx))
        if unique:
            keys = [i[0] for i in keys]
        kwds['iterate_key'] = True

        # Call the base class constructor.
        CombinatorialFreeModule.__init__(self, algebra,
                                         basis_keys=keys,
                                         category=category,
                                         names=names,
                                         **kwds)

    Element = FreeGradedModuleElement

    def change_ring(self, algebra):
        r"""
        Change the base ring of ``self``.

        INPUT:

        - ``algebra`` -- a connected graded algebra

        OUTPUT:

        The free graded module over ``algebra`` defined with the same
        number of generators of the same degrees as ``self``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: A2 = SteenrodAlgebra(2, profile=(3,2,1))

            sage: M = FreeGradedModule(A, [0,1])
            sage: N = M.change_ring(A2); N
            Free graded left module on 2 generators over sub-Hopf algebra of
             mod 2 Steenrod algebra, milnor basis, profile function [3, 2, 1]

        Changing back yields the original module::

            sage: N.change_ring(A) is M
            True
        """
        # We use the base class to avoid the category mixed one
        return type(self).__base__(algebra, self.generator_degrees(),
                                   prefix=self.prefix(), names=self._names)

    def _repr_(self):
        r"""
        Construct a string representation of ``self``.

        TESTS::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (0,2,4))
            sage: M
            Free graded left module on 3 generators over
             mod 2 Steenrod algebra, milnor basis
        """
        return ("Free graded left module on %s generator%s over %s"
                % (len(self._generator_degrees),
                   "" if len(self._generator_degrees) == 1 else "s",
                   self.base_ring()))

    def generator_degrees(self):
        r"""
        The degrees of the module generators.

        OUTPUT:

        A tuple containing the degrees of the generators for this
        module, in the order that the generators were given when
        ``self`` was constructed.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (-2,2,4))
            sage: M.generator_degrees()
            (-2, 2, 4)
        """
        return self._generator_degrees

    def is_trivial(self) -> bool:
        r"""
        Return ``True`` if this module is trivial and ``False`` otherwise.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: FreeGradedModule(A, (-2,2,4)).is_trivial()
            False
            sage: FreeGradedModule(A, ()).is_trivial()
            True
        """
        return not self._generator_degrees

    def connectivity(self):
        r"""
        The connectivity of ``self``.

        OUTPUT:

        An integer equal to the minimal degree of all the generators, if
        this module is non-trivial.  Otherwise, `+\infty`.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (-2,2,4))
            sage: M.connectivity()
            -2

        TESTS::

            sage: M = FreeGradedModule(A, ())
            sage: M.is_trivial()
            True
            sage: M.connectivity()
            +Infinity
        """
        return min(self.generator_degrees() + (infinity,))

    def _element_constructor_(self, coefficients):
        r"""
        Construct any element of ``self``.

        INPUT:

        - ``coefficients`` -- tuple of coefficient (i.e. elements of the
          algebra for this module), an element of FreeGradedModule, or the
          zero integer constant

        OUTPUT:

        An instance of the element class with coefficients from
        ``coefficients``, the element ``coefficients`` if it already
        was an element, or the zero module element.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M.<a,b,c> = FreeGradedModule(A, (0,2,4))

            sage: zero = M(0); zero
            0

            sage: e = M((Sq(4), Sq(2), 1)); e
            Sq(4)*a + Sq(2)*b + c

            sage: e is M(e)
            True
        """
        if isinstance(coefficients, self.element_class):
            return coefficients

        if not coefficients:
            return self.zero()

        A = self.base_ring()
        return self._from_dict({b: A(c) for (c, b) in zip(coefficients, self._indices) if c},
                               remove_zeros=False)

    def an_element(self, n=None):
        r"""
        Return an element of ``self``.

        This function chooses deterministically an element of the module
        in the given degree.

        INPUT:

        - ``n`` -- (optional) the degree of the element to construct

        OUTPUT:

        An element (of the given degree if specified).

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (0,2,4))
            sage: M.an_element(172)
            Sq(0,0,2,0,1,0,1)*g[0] + Sq(0,4,0,0,1,0,1)*g[2] + Sq(7,1,0,0,1,0,1)*g[4]

        Zero is the only element in the trivial module::

            sage: FreeGradedModule(A, ()).an_element()
            0
        """
        if not self._generator_degrees:
            return self.zero()

        if n is None:
            n = max(self.generator_degrees()) + 7

        coefficients = []
        for g in self.generator_degrees():
            basis = self.base_ring().basis(n - g) if n >= g else ()
            # All of the algebra generators in basis will bring the
            # module generator in dimension g to dimension
            # g + (topDimension - g) = topDimension.  Picking any one of them
            # will do, so we pick the one with index (g (mod l)).
            l = len(basis)
            if l:
                coefficients.append(basis[g % l])
            else:
                coefficients.append(self.base_ring().zero())

        return self(coefficients)

    # @cached_method
    def basis_elements(self, n):
        r"""
        Return a basis for the free module of degree ``n`` module elements.

        .. NOTE::

            Suppose ``self`` is a module over the graded algebra `A` with
            base ring `R`. This returns a basis as a free module over `R`,
            not a basis as a free module over `A`.

        INPUT:

        - ``n`` -- integer

        OUTPUT:

        A sequence of homogeneous module elements of degree ``n``, which
        is a basis for the free module of all degree ``n`` module elements.

        .. SEEALSO::

            :meth:`vector_presentation`, :meth:`element_from_coordinates`

        EXAMPLES::

            sage: A = SteenrodAlgebra(2)
            sage: M.<m0, m2, m4> = A.free_graded_module((0,2,4))
            sage: M.basis_elements(8)
            (Sq(1,0,1)*m0,
             Sq(2,2)*m0,
             Sq(5,1)*m0,
             Sq(8)*m0,
             Sq(0,2)*m2,
             Sq(3,1)*m2,
             Sq(6)*m2,
             Sq(1,1)*m4,
             Sq(4)*m4)
        """
        return tuple([self.term(self._indices[i], coeff)
                      for i in range(len(self._generator_degrees))
                      for coeff in self._basis_coeffs(n, i)])

    def _basis_coeffs(self, d, i):
        r"""
        Return a basis for the free module of degree ``d`` module elements
        corresponding to the ``i``-th generator.

        .. NOTE::

            Suppose ``self`` is a module over the graded algebra `A` with
            base ring `R`. This returns a basis as a free module over `R`,
            not a basis as a free module over `A`.

        INPUT:

        - ``d`` -- integer; the degree
        - ``i`` -- integer; the factor

        EXAMPLES::

            sage: A = SteenrodAlgebra(2)
            sage: M = A.free_graded_module((0,1))
            sage: M._basis_coeffs(3, 0)
            (Sq(0,1), Sq(3))
            sage: M._basis_coeffs(3, 1)
            (Sq(2),)
        """
        return self._cached_basis_coeffs(d - self._generator_degrees[i])

    @cached_method
    def _cached_basis_coeffs(self, d):
        """
        Return the basis for the degree ``d`` part of the base algebra.

        EXAMPLES::

            sage: E.<x,y> = ExteriorAlgebra(QQ)
            sage: M = E.free_graded_module((0,0,1))
            sage: M._cached_basis_coeffs(0)
            (1,)
            sage: M._cached_basis_coeffs(1)
            (x, y)
            sage: M._cached_basis_coeffs(2)
            (x*y,)
            sage: M._cached_basis_coeffs(3)
            ()
        """
        return tuple(self.base_ring().basis(d))

    @cached_method
    def element_from_coordinates(self, coordinates, n):
        r"""
        The module element of degree ``n`` having the given coordinates
        with respect to the basis of module elements given by
        :meth:`basis_elements`.

        INPUT:

        - ``coordinates`` -- a sequence of elements of the ground ring
        - ``n`` -- integer

        OUTPUT: a module element of degree ``n``

        .. SEEALSO::

            :meth:`vector_presentation`, and :meth:`basis_elements`.

        EXAMPLES::

            sage: A = SteenrodAlgebra(2)
            sage: M = A.free_graded_module((0,1))
            sage: x = M.element_from_coordinates((0,1,0,1), 5); x
            Sq(5)*g[0] + Sq(4)*g[1]
            sage: basis = M.basis_elements(5)
            sage: y = 0*basis[0] + 1*basis[1] + 0*basis[2] + 1*basis[3]
            sage: x == y
            True

            sage: M.element_from_coordinates((0,0,0,0), 5)
            0
        """
        D = self.vector_presentation(n).dimension()
        if len(coordinates) != D:
            raise ValueError('the given coordinate vector has incorrect length (%d); '
                             'it should have length %d' % (len(coordinates), D))

        # Performance testing using this real life example:
        #
        # sage: A = SteenrodAlgebra(2)
        # sage: from sage.modules.fp_graded.steenrod.module import SteenrodFPModule
        # sage: rels = [[Sq(1),0,0,0], [Sq(2),0,0,0], [Sq(4),0,0,0], [Sq(8),0,0,0],
        # ....:         [0,Sq(1),0,0], [0,Sq(2),0,0], [0,Sq(4),0,0], [Sq(31),Sq(14),0,0],
        # ....:         [0,Sq(20),0,0], [0,0,Sq(1),0], [0,0,Sq(2),0], [0,Sq(31),Sq(6),0],
        # ....:         [0,0,Sq(8),0], [0,0,0,Sq(1)], [0,0,Sq(31),Sq(2)], [0,0,0,Sq(4)], [0,0,0,Sq(8)] ]
        # sage: M = SteenrodFPModule(A, [0, 17, 42, 71], relations=rels)
        # sage: res = M.resolution(2, top_dim=30, verbose=True)
        #
        # This function was called a total of 2897 times during the computation,
        # and the total running time of the entire computation dropped from
        # 57 to 21 seconds by adding the optimization.

        ret = {}
        A = self.base_ring()
        j = 0
        for i, key in enumerate(self._indices):
            B = self._basis_coeffs(n, i)
            coeff = A.linear_combination((b, coordinates[j + ind])
                                         for ind, b in enumerate(B))
            if coeff:
                ret[key] = coeff
            j += len(B)

        if not ret:
            return self.zero()
        return self.element_class(self, ret)

    @cached_method
    def vector_presentation(self, n):
        r"""
        Return a free module over the ground ring of the module algebra
        isomorphic to the degree ``n`` elements of ``self``.

        Let `\mathcal{k}` be the ground ring of the algebra over this module
        is defined, and let `M_n` be the free module of module elements of
        degree ``n``.

        The return value of this function is the free module
        `\mathcal{k}^{r}` where `r = dim(M_n)`.

        The isomorphism between `k^{r}` and `M_n` is given by the
        bijection taking the standard basis element `e_i` to the `i`-th
        element of the array returned by :meth:`basis_elements`.

        INPUT:

        - ``n`` -- integer degree

        OUTPUT:

        A free module over the ground ring of the algebra over which
        ``self`` is defined, isomorphic to the free module of module
        elements of degree ``n``.

        .. SEEALSO::

            :meth:`basis_elements`, :meth:`element_from_coordinates`

        EXAMPLES::

            sage: A1 = SteenrodAlgebra(2, profile=[2,1])
            sage: M.<x> = A1.free_graded_module((0,))
            sage: M.vector_presentation(3)
            Vector space of dimension 2 over Finite Field of size 2
            sage: M.basis_elements(3)
            (Sq(0,1)*x, Sq(3)*x)
            sage: [M.vector_presentation(i).dimension() for i in range(-2, 9)]
            [0, 0, 1, 1, 1, 2, 1, 1, 1, 0, 0]

        TESTS::

            sage: A = SteenrodAlgebra(2)
            sage: M = A.free_graded_module((0,2,4))
            sage: V = M[4]; V
            Vector space of dimension 4 over Finite Field of size 2
            sage: V.dimension()
            4
        """
        m = len(self._generator_degrees)
        return FreeModule(self.base_ring().base_ring(), sum(len(self._basis_coeffs(n, i))
                                                            for i in range(m)))

    __getitem__ = vector_presentation

    def generator(self, index):
        r"""
        Return the module generator with the given index.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (0,2,4))
            sage: M.generator(0)
            g[0]
            sage: M.generator(1)
            g[2]
            sage: M.generator(2)
            g[4]
        """
        try:
            return self.gens()[index]
        except IndexError:
            raise ValueError('the parent module has generators in the index '
                             'range [0, %s]; generator %s does not exist' %
                             (len(self.generator_degrees()) - 1, index))

    gen = generator

    def generators(self):
        r"""
        Return all the module generators.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (-2,1))
            sage: M.generators()
            (g[-2], g[1])
        """
        return self.gens()

    def _Hom_(self, Y, category):
        r"""
        The internal hook used by the free function
        :meth:`sage.categories.homset.hom.Hom` to create homsets
        involving ``self``.

        TESTS::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (0,1))
            sage: M._Hom_(M, category=None)
            Set of Morphisms from Free graded left module on 2 generators
              over mod 2 Steenrod algebra, milnor basis
             to Free graded left module on 2 generators
              over mod 2 Steenrod algebra, milnor basis
             in Category of finite dimensional graded modules with basis
              over mod 2 Steenrod algebra, milnor basis

            sage: from sage.modules.fp_graded.module import FPModule
            sage: F = FPModule(A, [1,3])
            sage: Hom(M, F)
            Set of Morphisms from Free graded left module on 2 generators ...
        """
        from .free_homspace import FreeGradedModuleHomspace
        return FreeGradedModuleHomspace(self, Y, category)

    def suspension(self, t):
        r"""
        Suspend ``self`` by the given degree ``t``.

        INPUT:

        - ``t`` -- integer

        OUTPUT:

        A module which is isomorphic to this module by a shift
        of degrees by the integer ``t``.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: M = FreeGradedModule(A, (0,2,4))
            sage: M.suspension(4).generator_degrees()
            (4, 6, 8)
            sage: M.suspension(-4).generator_degrees()
            (-4, -2, 0)
        """
        degs = tuple(g + t for g in self.generator_degrees())
        return FreeGradedModule(algebra=self.base_ring(),
                                generator_degrees=degs)

    def has_relations(self) -> bool:
        r"""
        Return ``False`` as this has no relations.

        This is for compatibility with
        :class:`~sage.modules.fp_graded.module.FPModule`.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: F = FreeGradedModule(A, (-2,2,4))
            sage: F.has_relations()
            False
        """
        return False

    def relations(self):
        r"""
        Return the relations of ``self``, which is ``()``.

        This is for compatibility with
        :class:`~sage.modules.fp_graded.module.FPModule`.

        EXAMPLES::

            sage: from sage.modules.fp_graded.free_module import FreeGradedModule
            sage: A = SteenrodAlgebra(2)
            sage: F = FreeGradedModule(A, (-2,2,4))
            sage: F.relations()
            ()
        """
        return ()

    def resolution(self, k, top_dim=None, verbose=False):
        r"""
        Return a free resolution of ``self`` of length ``k``.

        Since ``self`` is free, the initial map in the resolution will
        be the identity, and the rest of the maps will be zero.

        INPUT:

        - ``k`` -- nonnegative integer
        - ``top_dim`` -- stop the computation at this degree. Ignored,
          for compatibility with
          :meth:`sage.modules.fp_graded.module.FPModule.resolution`.
        - ``verbose`` -- boolean (default: ``False``); controls whether log
          messages should be emitted

        OUTPUT:

        A list of homomorphisms `[1_M, 0, 0, \ldots, 0]` consisting of
        the identity map on this module followed by zero maps. Other
        than this module, the other modules in the resolution will be
        zero.

        EXAMPLES::

            sage: E.<x,y,z> = ExteriorAlgebra(QQ)
            sage: M = E.free_graded_module((1,2))
            sage: M.resolution(0)
            [Module endomorphism of Free graded left module on 2 generators over The exterior algebra of rank 3 over Rational Field
               Defn: g[1] |--> g[1]
                     g[2] |--> g[2]]
            sage: M.resolution(1)
            [Module endomorphism of Free graded left module on 2 generators over The exterior algebra of rank 3 over Rational Field
               Defn: g[1] |--> g[1]
                     g[2] |--> g[2],
             Module morphism:
               From: Free graded left module on 0 generators over The exterior algebra of rank 3 over Rational Field
               To:   Free graded left module on 2 generators over The exterior algebra of rank 3 over Rational Field]
            sage: M.resolution(4)
            [Module endomorphism of Free graded left module on 2 generators over The exterior algebra of rank 3 over Rational Field
               Defn: g[1] |--> g[1]
                     g[2] |--> g[2],
             Module morphism:
               From: Free graded left module on 0 generators over The exterior algebra of rank 3 over Rational Field
               To:   Free graded left module on 2 generators over The exterior algebra of rank 3 over Rational Field,
             Module endomorphism of Free graded left module on 0 generators over The exterior algebra of rank 3 over Rational Field,
             Module endomorphism of Free graded left module on 0 generators over The exterior algebra of rank 3 over Rational Field,
             Module endomorphism of Free graded left module on 0 generators over The exterior algebra of rank 3 over Rational Field]
        """
        if k < 0:
            raise ValueError('the length of the resolution must be nonnegative')

        # The first map \epsilon is the identity map
        ret_complex = [Hom(self, self).identity()]

        if k == 0:
            return ret_complex

        # All subsequent maps are trivial since self is free
        T = self.base_ring().free_graded_module(())
        ret_complex.append(Hom(T, self).zero())

        if k == 1:
            return ret_complex

        return ret_complex + [Hom(T, T).zero()] * (k - 1)

    def minimal_presentation(self, top_dim=None, verbose=False):
        r"""
        Return a minimal presentation of ``self``.

        OUTPUT: the identity morphism as ``self`` is free

        EXAMPLES::

            sage: from sage.modules.fp_graded.module import FPModule
            sage: A2 = SteenrodAlgebra(2)

            sage: M = A2.free_graded_module([0,1])
            sage: M.minimal_presentation().is_identity()
            True
        """
        return Hom(self, self).identity()



"""
Complex plots

AUTHORS:

- Robert Bradshaw (2009): initial version
- David Lowry-Duda (2022): incorporate matplotlib colormaps
"""
# ****************************************************************************
#       Copyright (C) 2009 Robert Bradshaw <robertwb@math.washington.edu>,
#       Copyright (C) 2022 David Lowry-Duda <david@lowryduda.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on, sig_off, sig_check


# Note: don't import matplotlib at module level. It takes a surprisingly
#       long time to load!
cimport numpy as cnumpy

from sage.plot.primitive import GraphicPrimitive
from sage.misc.decorators import options
from sage.rings.complex_double cimport ComplexDoubleElement
from sage.arith.srange import srange

from libc.math cimport hypot, atan2, atan, log, pow, sqrt
from sage.arith.constants cimport M_PI as PI

from sage.libs.gsl.complex cimport *


DEFAULT_LOGARITHMIC_CONTOUR_BASE = 2
DEFAULT_LINEAR_CONTOUR_BASE = 10


cdef inline ComplexDoubleElement new_CDF_element(double x, double y):
    z = <ComplexDoubleElement>ComplexDoubleElement.__new__(ComplexDoubleElement)
    GSL_SET_COMPLEX(&z._complex, x, y)
    return z


cdef inline double mag_to_lightness(double r, double rate=0.5) noexcept:
    """
    Return a lightness for the given magnitude.

    Small magnitudes are darker and large magnitudes are lighter, and the
    lightness smoothly transitions.

    Tweak the rate to adjust how the magnitude affects the color. For
    instance, changing ``rate`` to `1` will cause anything near zero to be
    much darker and poles to be much lighter, while `0.25` would cause the
    reverse effect.

    INPUT:

    - ``r`` -- a nonnegative real number; the magnitude

    - ``rate`` -- a positive real number; how quickly changes in magnitude
      affect changes in output lightness

    OUTPUT:

    A value between `-1` (black) and `+1` (white), inclusive.

    .. SEEALSO::

        - :func:`sage.plot.complex_plot.cyclic_logarithmic_mag_to_lightness`
        - :func:`sage.plot.complex_plot.mag_and_arg_to_lightness`

    EXAMPLES:

    This tests it implicitly::

        sage: from sage.plot.complex_plot import complex_to_rgb
        sage: complex_to_rgb([[0, 1, 10]])  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.77172568, 0.        , 0.        ],
                [1.        , 0.22134776, 0.22134776]]])

    Changing ``rate`` changes the rate of growth::

        sage: complex_to_rgb([[0, 1, 10]], dark_rate=1)  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.77172568, 0.        , 0.        ],
                [1.        , 0.49693961, 0.49693961]]])
    """
    if rate == 0.5:
        return atan(log(sqrt(r)+1)) * (4/PI) - 1
    else:
        return atan(log(pow(r, rate)+1)) * (4/PI) - 1


cdef inline double cyclic_logarithmic_mag_to_lightness(double r, double base=2) noexcept:
    r"""
    Return a lightness for the given magnitude.

    This modifies the lightness around magnitudes of size `base^n` for integer
    `n` to create the appearance of logarithmically spaced contours.

    INPUT:

    - ``r`` -- a nonnegative real number; the magnitude

    - ``base`` -- a real number (default: `2`); contours will appear at integer
      powers of ``base``. This should be greater than `1`.

    OUTPUT:

    A value between `-1` (black) and `+1` (white), inclusive.

    .. SEEALSO::

        - :func:`sage.plot.complex_plot.mag_to_lightness`
        - :func:`sage.plot.complex_plot.mag_and_arg_to_lightness`
        - :func:`sage.plot.complex_plot.cyclic_linear_mag_to_lightness`

    EXAMPLES:

    This tests it implicitly::

        sage: from sage.plot.complex_plot import complex_to_rgb
        sage: complex_to_rgb([[0, 1, 10]], contoured=True,  # abs tol 1e-4
        ....:                contour_type='logarithmic')
        array([[[1.        , 0.        , 0.        ],
                [1.        , 0.15      , 0.15      ],
                [0.98903595, 0.        , 0.        ]]])

    We set contours to be multiples of `5` apart::

        sage: complex_to_rgb([[0, 1, 10]], contoured=True,  # abs tol 1e-4
        ....:                contour_type='logarithmic', contour_base=5)
        array([[[1.        , 0.        , 0.        ],
                [1.        , 0.15      , 0.15      ],
                [0.93466172, 0.        , 0.        ]]])
    """
    if r < 1e-10:
        return 0.0
    rem  = (log(r) / log(base)) % 1
    if rem < 0:  # Choose positive mod representative
        rem += 1
    return .15 - rem/2.


cdef inline double cyclic_linear_mag_to_lightness(double r, double base=10) noexcept:
    r"""
    Return a lightness for the given magnitude.

    This modifies the lightness around magnitudes of size `n*base` for integer
    `n` to create the appearance of linearly spaced contours.

    INPUT:

    - ``r`` -- a nonnegative real number; the magnitude

    - ``base`` -- a positive real number (default: `10`); contours will appear
      at integer multiples of ``base``

    OUTPUT:

    A value between `-1` (black) and `+1` (white), inclusive.

    .. SEEALSO::

        - :func:`sage.plot.complex_plot.mag_to_lightness`
        - :func:`sage.plot.complex_plot.mag_and_arg_to_lightness`
        - :func:`sage.plot.complex_plot.cyclic_logarithmic_mag_to_lightness`

    EXAMPLES:

    This tests it implicitly::

        sage: from sage.plot.complex_plot import complex_to_rgb
        sage: complex_to_rgb([[1, 5, 11]], contoured=True,  # abs tol 1e-4
        ....:                contour_type='linear')
        array([[[1. , 0.1, 0.1],
                [0.9, 0. , 0. ],
                [1. , 0.1, 0.1]]])

    In the above example, note that both `1` and `11` have the same imaginary
    part and are precisely `10` (the default contour separation) apart. If we
    set contours to be multiples of `3` apart, the values are no longer the
    same, but the values for `5` and `11` should be::

        sage: complex_to_rgb([[1, 5, 11]], contoured=True,  # abs tol 1e-4
        ....:                contour_type='linear', contour_base=3)
        array([[[0.98333333, 0.        , 0.        ],
                [0.81666667, 0.        , 0.        ],
                [0.81666667, 0.        , 0.        ]]])
    """
    rem = (r / base) % 1
    if rem < 0:  # Choose positive mod representative
        rem += 1
    return .15 - rem/2.


cdef inline double mag_and_arg_to_lightness(double r, double arg,
                                            double base=2, int nphases=10) noexcept:
    r"""
    Return a lightness for the given magnitude and argument.

    This modifies the lightness around magnitudes of size ``base^n`` for
    integer `n`, and also around arguments of the form `(2 \pi) / nphases * n` for
    integer `n`. This creates the appearance of tiles, bounded by
    (base ``base``) logarithmically spaced magnitude contours and by
    (``nphases`` many) evenly spaced argument contours.

    INPUT:

    - ``r`` -- a nonnegative real number

    - ``arg`` -- a real number

    - ``base`` -- a positive real number (default: ``2``); contours will appear
      at integer powers of ``base``. This should be greater than `1`.

    - ``nphases`` -- positive integer (default: `10`); how many phase
      contours to represent a change of argument by `2 \pi`

    OUTPUT:

    A value between `-1` (black) and `+1` (white), inclusive.

    .. SEEALSO::

        - :func:`sage.plot.complex_plot.mag_to_lightness`
        - :func:`sage.plot.complex_plot.cyclic_logarithmic_mag_to_lightness`

    EXAMPLES:

    This tests it implicitly::

        sage: from sage.plot.complex_plot import complex_to_rgb
        sage: complex_to_rgb([[0, 1, 10]], tiled=True)  # abs tol 1e-4
        array([[[1.        , 0.        , 0.        ],
                [1.        , 0.15      , 0.15      ],
                [1.        , 0.06951798, 0.06951798]]])
        sage: complex_to_rgb([[0, 1 + 1j, -3 - 5j]], tiled=True)  # abs tol 1e-4
        array([[[1.        , 0.        , 0.        ],
                [0.9625    , 0.721875  , 0.        ],
                [0.        , 0.01371897, 0.85409323]]])

    Adjusting the tiling parameters should create relatively small
    differences::

        sage: complex_to_rgb([[0, 1 + 1j, -3 - 5j]],  # abs tol 1e-4
        ....:                tiled=True, nphases=15)
        array([[[1.        , 0.        , 0.        ],
                [0.80625   , 0.6046875 , 0.        ],
                [0.        , 0.01243417, 0.77410628]]])
        sage: complex_to_rgb([[0, 1 + 1j, -3 - 5j]],  # abs tol 1e-4
        ....:                tiled=True, contour_base=5, nphases=15)
        array([[[1.        , 0.        , 0.        ],
                [0.87741543, 0.65806157, 0.        ],
                [0.        , 0.01423401, 0.88615776]]])
    """
    if r < 1e-10:
        return 0.0
    cdef double r_rem, arg_rem
    r_rem  = (log(r) / log(base)) % 1
    arg_rem = (nphases * arg / (2*PI)) % 1
    if r_rem < 0:  # Choose positive mod representatives
        r_rem += 1
    if arg_rem < 0:
        arg_rem += 1
    return 0.15 - (r_rem)/4. - (arg_rem)/4.


def complex_to_rgb(z_values, contoured=False, tiled=False,
                   contour_type='logarithmic', contour_base=None,
                   dark_rate=0.5, nphases=10):
    r"""
    Convert a grid of complex numbers to a grid of rgb values using a default
    choice of colors.

    INPUT:

    - ``z_values`` -- a grid of complex numbers, as a list of lists

    - ``contoured`` -- boolean (default: ``False``); causes magnitude to be
      indicated through contour-like adjustments to lightness

    - ``tiled`` -- boolean (default: ``False``); causes magnitude and argument to
      be indicated through contour-like adjustments to lightness

    - ``nphases`` -- positive integer (default: 10); when ``tiled=True``,
      this is the number of divisions the phase is divided into

    - ``contour_type`` -- either ``'logarithmic'``, or ``'linear'`` (default:
      ``'logarithmic'``); causes added contours to be of given type when
      ``contoured=True``.

    - ``contour_base`` -- positive integer; when ``contour_type`` is
      ``'logarithmic'``, this sets logarithmic contours at multiples of
      ``contour_base`` apart. When ``contour_type`` is ``'linear'``, this sets
      contours at distances of ``contour_base`` apart. If ``None``, then a
      default is chosen depending on ``contour_type``.

    - ``dark_rate`` -- a positive number (default: 0.5); affects how quickly
      magnitudes affect how light/dark the image is. When there are contours,
      this affects how visible each contour is. Large values (near `1.0`) have
      very strong, immediate effects, while small values (near `0.0`) have
      gradual effects.

    OUTPUT:

    An `N \times M \times 3` floating point Numpy array ``X``, where
    ``X[i,j]`` is an (r,g,b) tuple.

    .. SEEALSO::

        :func:`sage.plot.complex_plot.complex_to_cmap_rgb`

    EXAMPLES:

    We can call this on grids of complex numbers::

        sage: from sage.plot.complex_plot import complex_to_rgb
        sage: complex_to_rgb([[0, 1, 1000]])  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.77172568, 0.        , 0.        ],
                [1.        , 0.64421177, 0.64421177]]])
        sage: complex_to_rgb([[0, 1j, 1000j]])  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.38586284, 0.77172568, 0.        ],
                [0.82210588, 1.        , 0.64421177]]])
        sage: complex_to_rgb([[0, 1, 1000]], contoured=True)   # abs tol 1e-4
        array([[[1.        , 0.        , 0.        ],
                [1.        , 0.15      , 0.15      ],
                [0.66710786, 0.        , 0.        ]]])
        sage: complex_to_rgb([[0, 1, 1000]], tiled=True)   # abs tol 1e-4
        array([[[1.        , 0.        , 0.        ],
                [1.        , 0.15      , 0.15      ],
                [0.90855393, 0.        , 0.        ]]])

    We can change contour types and the distances between contours::

        sage: complex_to_rgb([[0, 1 + 1j, 3 + 4j]],  # abs tol 1e-4
        ....:                contoured=True, contour_type='logarithmic', contour_base=3)
        array([[[1.        , 0.        , 0.        ],
                [0.99226756, 0.74420067, 0.        ],
                [0.91751324, 0.81245954, 0.        ]]])
        sage: complex_to_rgb([[0, 1 + 1j, 3 + 4j]],  # abs tol 1e-4
        ....:                contoured=True, contour_type='linear', contour_base=3)
        array([[[1.        , 0.15      , 0.15      ],
                [0.91429774, 0.6857233 , 0.        ],
                [0.81666667, 0.72315973, 0.        ]]])

    Lowering ``dark_rate`` causes colors to go to black more slowly near `0`::

        sage: complex_to_rgb([[0, 0.5, 1]], dark_rate=0.4)  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.65393731, 0.        , 0.        ],
                [0.77172568, 0.        , 0.        ]]])
        sage: complex_to_rgb([[0, 0.5, 1]], dark_rate=0.2)  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.71235886, 0.        , 0.        ],
                [0.77172568, 0.        , 0.        ]]])
    """
    import numpy as np

    cdef unsigned int i, j, imax, jmax
    cdef double x, y, mag, arg
    cdef double lightness, hue, top, bot
    cdef double r, g, b
    cdef int ihue
    cdef ComplexDoubleElement z
    from sage.rings.complex_double import CDF

    imax = len(z_values)
    jmax = len(z_values[0])
    cdef cnumpy.ndarray[cnumpy.float_t, ndim=3, mode='c'] rgb = np.empty(dtype=float, shape=(imax, jmax, 3))

    if contour_base is None:
        if contour_type == "linear":
            contour_base = DEFAULT_LINEAR_CONTOUR_BASE
        else:
            contour_base = DEFAULT_LOGARITHMIC_CONTOUR_BASE
    if contour_type not in ("linear", "logarithmic"):
        raise ValueError("Unrecognized contour type argument {}.".format(contour_type))
    if contour_base <= 0:
        raise ValueError("contour_base must be positive")

    sig_on()
    for i in range(imax):
        row = z_values[i]
        for j in range(jmax):
            zz = row[j]
            if type(zz) is ComplexDoubleElement:
                z = <ComplexDoubleElement>zz
            else:
                z = CDF(zz)
            x = GSL_REAL(z._complex)
            y = GSL_IMAG(z._complex)
            mag = hypot(x, y)
            arg = atan2(y, x) # math module arctan has range from -pi to pi, so cut along negative x-axis

            if tiled:
                lightness = mag_and_arg_to_lightness(
                    mag, arg, base=contour_base, nphases=nphases
                )
            elif contoured:
                if contour_type == "logarithmic":
                    lightness = cyclic_logarithmic_mag_to_lightness(mag, base=contour_base)
                else:
                    lightness = cyclic_linear_mag_to_lightness(mag, base=contour_base)
            else:
                lightness = mag_to_lightness(mag, rate=dark_rate)

            if lightness < 0: # in hsv, variable value, full saturation (s=1, v=1+lightness)
                bot = 0
                top = (1+lightness)
            else: # in hsv, variable saturation, full value (v=1, s=1-lightness)
                bot = lightness
                top = 1

            # Note that does same thing as colorsys module hsv_to_rgb
            # for this setup, but in Cython
            hue = 3*arg/PI

            if hue < 0:
                hue += 6  # usual hsv hue is thus h=arg/(2*pi) for positive, h=arg/(2*PI)+1 for negative
            ihue = <int>hue
            if ihue == 0:
                r = top
                g = bot + hue * (top-bot)
                b = bot
            elif ihue == 1:
                r = bot + (2-hue) * (top-bot)
                g = top
                b = bot
            elif ihue == 2:
                r = bot
                g = top
                b = bot + (hue-2) * (top-bot)
            elif ihue == 3:
                r = bot
                g = bot + (4-hue) * (top-bot)
                b = top
            elif ihue == 4:
                r = bot + (hue-4) * (top-bot)
                g = bot
                b = top
            else:
                r = top
                g = bot
                b = bot + (6-hue) * (top-bot)

            rgb[i, j, 0] = r
            rgb[i, j, 1] = g
            rgb[i, j, 2] = b

    sig_off()
    nan_indices = np.isnan(rgb).any(-1)     # Mask for undefined points
    rgb[nan_indices] = 1                    # Make nan_indices white
    return rgb


def complex_to_cmap_rgb(z_values, cmap='turbo', contoured=False, tiled=False,
                        contour_type='logarithmic', contour_base=None,
                        dark_rate=0.5, nphases=10):
    r"""
    Convert a grid of complex numbers to a grid of rgb values using colors
    taken from given colormap.

    INPUT:

    - ``z_values`` -- a grid of complex numbers, as a list of lists

    - ``cmap`` -- the string name of a matplotlib colormap, or an instance
      of a matplotlib Colormap (default: ``'turbo'``)

    - ``contoured`` -- boolean (default: ``False``); causes magnitude to be
      indicated through contour-like adjustments to lightness

    - ``tiled`` -- boolean (default: ``False``); causes magnitude and argument to
      be indicated through contour-like adjustments to lightness

    - ``nphases`` -- positive integer (default: 10); when ``tiled=True``,
      this is the number of divisions the phase is divided into

    - ``contour_type`` -- either ``'logarithmic'``, or ``'linear'`` (default:
      ``'logarithmic'``); causes added contours to be of given type when
      ``contoured=True``.

    - ``contour_base`` -- positive integer; when ``contour_type`` is
      ``'logarithmic'``, this sets logarithmic contours at multiples of
      ``contour_base`` apart. When ``contour_type`` is ``'linear'``, this sets
      contours at distances of ``contour_base`` apart. If ``None``, then a
      default is chosen depending on ``contour_type``.

    - ``dark_rate`` -- a positive number (default: 0.5); affects how quickly
      magnitudes affect how light/dark the image is. When there are contours,
      this affects how visible each contour is. Large values (near 1.0) have
      very strong, immediate effects, while small values (near 0.0) have
      gradual effects.

    OUTPUT:

    An `N \times M \times 3` floating point Numpy array ``X``, where
    ``X[i,j]`` is an (r, g, b) tuple.

    .. SEEALSO::

        :func:`sage.plot.complex_plot.complex_to_rgb`

    EXAMPLES:

    We can call this on grids of complex numbers::

        sage: from sage.plot.complex_plot import complex_to_cmap_rgb
        sage: complex_to_cmap_rgb([[0, 1, 1000]])  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.49669808, 0.76400071, 0.18024425],
                [0.87320419, 0.99643856, 0.72730967]]])
        sage: complex_to_cmap_rgb([[0, 1, 1000]], cmap='viridis')  # abs tol 1e-4
        array([[[0.        , 0.        , 0.        ],
                [0.0984475 , 0.4375291 , 0.42487821],
                [0.68959896, 0.84592555, 0.84009311]]])


    We can change contour types and the distances between contours::

        sage: complex_to_cmap_rgb([[0, 1 + 1j, 3 + 4j]], contoured=True,  # abs tol 1e-4
        ....:                     contour_type='logarithmic', contour_base=3)
        array([[[0.64362   , 0.98999   , 0.23356   ],
                [0.93239357, 0.81063338, 0.21955399],
                [0.95647342, 0.74861225, 0.14963982]]])
        sage: complex_to_cmap_rgb([[0, 1 + 1j, 3 + 4j]], cmap='turbo',   # abs tol 1e-4
        ....:                     contoured=True, contour_type='linear', contour_base=3)
        array([[[0.71246796, 0.9919238 , 0.3816262 ],
                [0.92617785, 0.79322304, 0.14779989],
                [0.95156284, 0.72025117, 0.05370383]]])

    We see that changing ``dark_rate`` affects how visible contours are. In this
    example, we set ``contour_base=5`` and note that the points `0` and `1 + i`
    are far away from contours, but `2.9 + 4i` is near (and just below) a
    contour. Raising ``dark_rate`` should have strong effects on the last
    coloration and weaker effects on the others::

        sage: complex_to_cmap_rgb([[0, 1 + 1j, 2.9 + 4j]], cmap='turbo',  # abs tol 1e-4
        ....:                     contoured=True, dark_rate=0.05, contour_base=5)
        array([[[0.64362   , 0.98999   , 0.23356   ],
                [0.93334746, 0.81330523, 0.23056563],
                [0.96357185, 0.75337736, 0.19440913]]])
        sage: complex_to_cmap_rgb([[0, 1 + 1j, 2.9 + 4j]], cmap='turbo',  # abs tol 1e-4
        ....:                     contoured=True, dark_rate=0.85, contour_base=5)
        array([[[0.64362   , 0.98999   , 0.23356   ],
                [0.93874682, 0.82842892, 0.29289564],
                [0.57778954, 0.42703289, 0.02612716]]])
    """
    import numpy as np
    import matplotlib as mpl

    if isinstance(cmap, str):
        cmap = mpl.colormaps[cmap]

    if contour_base is None:
        if contour_type == "linear":
            contour_base = DEFAULT_LINEAR_CONTOUR_BASE
        else:
            contour_base = DEFAULT_LOGARITHMIC_CONTOUR_BASE
    if contour_type not in ("linear", "logarithmic"):
        raise ValueError("Unrecognized contour type argument {}.".format(contour_type))
    if contour_base <= 0:
        raise ValueError("contour_base must be positive")

    cdef unsigned int i, j, imax, jmax
    cdef double x, y, mag, arg
    cdef double lightness_delta
    cdef ComplexDoubleElement z
    from sage.rings.complex_double import CDF

    imax = len(z_values)
    jmax = len(z_values[0])
    cdef cnumpy.ndarray[cnumpy.float_t, ndim=3, mode='c'] als = np.empty(dtype=float, shape=(imax, jmax, 2))

    sig_on()
    for i in range(imax):
        row = z_values[i]
        for j in range(jmax):
            zz = row[j]
            if type(zz) is ComplexDoubleElement:
                z = <ComplexDoubleElement>zz
            else:
                z = CDF(zz)
            x = GSL_REAL(z._complex)
            y = GSL_IMAG(z._complex)
            mag = hypot(x, y)
            arg = atan2(y, x) # math module arctan has range from -pi to pi, so cut along negative x-axis
            if tiled:
                lightness_delta = mag_and_arg_to_lightness(
                    mag, arg, base=contour_base, nphases=nphases
                )
            elif contoured:
                if contour_type == "logarithmic":
                    lightness_delta = cyclic_logarithmic_mag_to_lightness(mag, base=contour_base)
                else:
                    lightness_delta = cyclic_linear_mag_to_lightness(mag, base=contour_base)
            else:
                lightness_delta = mag_to_lightness(mag, rate=dark_rate)
            als[i, j, 0] = arg
            als[i, j, 1] = lightness_delta
    sig_off()

    args = als[:,:,0]
    nan_indices = np.isnan(als).any(-1)            # Mask for undefined points
    normalized_colors = cmap((args + PI) / (2 * PI)) # break on negative reals
    normalized_colors = normalized_colors[:,:,:3]  # discard alpha channel
    lightdeltas = als[:,:,1]

    if tiled or contoured:
        rgbs = add_contours_to_rgb(normalized_colors, lightdeltas, dark_rate=dark_rate)
    else:
        rgbs = add_lightness_smoothing_to_rgb(normalized_colors, lightdeltas)

    # Apply mask, making nan_indices white
    rgbs[nan_indices] = 1

    return rgbs


def add_lightness_smoothing_to_rgb(rgb, delta):
    r"""
    Return an rgb array from given array of colors and lightness adjustments.

    This smoothly adds lightness from black (when ``delta`` is `-1`) to white
    (when ``delta`` is `1`).

    Each input `(r, g, b)` is modified by ``delta`` to be lighter or darker
    depending on the size of ``delta``. When ``delta`` is `-1`, the output is
    black. When ``delta`` is `+1`, the output is white. Colors
    piecewise-linearly vary from black to the initial `(r, g, b)` to white.

    We assume that the ``delta`` values come from a function like
    :func:`sage.plot.complex_plot.mag_to_lightness`, which maps magnitudes to
    the range `[-1, +1]`.

    INPUT:

    - ``rgb`` -- a grid of length 3 tuples `(r, g, b)`, as an
      `N \times M \times 3` numpy array

    - ``delta`` -- a grid of values as an `N \times M` numpy array; these
      represent how much to change the lightness of each `(r, g, b)`. Values
      should be in `[-1, 1]`.

    OUTPUT:

    An `N \times M \times 3` floating point Numpy array ``X``, where
    ``X[i,j]`` is an (r, g, b) tuple.

    .. SEEALSO::

        - :func:`sage.plot.complex_plot.complex_to_rgb`
        - :func:`sage.plot.complex_plot.add_contours_to_rgb`

    EXAMPLES:

    We can call this on grids of values::

        sage: # needs numpy
        sage: import numpy as np
        sage: from sage.plot.complex_plot import add_lightness_smoothing_to_rgb
        sage: add_lightness_smoothing_to_rgb(  # abs tol 1e-4
        ....:     np.array([[[0, 0.25, 0.5]]]), np.array([[0.75]]))
        array([[[0.75  , 0.8125, 0.875 ]]])
        sage: add_lightness_smoothing_to_rgb(  # abs tol 1e-4
        ....:     np.array([[[0, 0.25, 0.5]]]), np.array([[0.75]]))
        array([[[0.75  , 0.8125, 0.875 ]]])
    """
    import numpy as np
    delta = delta[:,:,np.newaxis]
    delta_pos = delta > 0.0
    rgb = (1.0 - np.abs(delta))*(rgb - delta_pos) + delta_pos
    rgb = np.clip(rgb, 0.0, 1.0)
    return rgb


def add_contours_to_rgb(rgb, delta, dark_rate=0.5):
    r"""
    Return an rgb array from given array of `(r, g, b)` and `(delta)`.

    Each input `(r, g, b)` is modified by ``delta`` to be lighter or darker
    depending on the size of ``delta``. Negative ``delta`` values darken the
    color, while positive ``delta`` values lighten the pixel.

    We assume that the ``delta`` values come from a function like
    :func:`sage.plot.complex_plot.mag_to_lightness`, which maps magnitudes to
    the range `[-1, +1]`.

    INPUT:

    - ``rgb`` -- a grid of length 3 tuples `(r, g, b)`, as an `N \times M
      \times 3` numpy array

    - ``delta`` -- a grid of values as an `N \times M` numpy array; these
      represent how much to change the lightness of each `(r, g, b)`. Values
      should be in `[-1, 1]`.

    - ``dark_rate`` -- a positive number (default: 0.5); affects how
      strongly visible the contours appear

    OUTPUT:

    An `N \times M \times 3` floating point Numpy array ``X``, where
    ``X[i,j]`` is an (r, g, b) tuple.

    .. SEEALSO::

        - :func:`sage.plot.complex_plot.complex_to_rgb`,
        - :func:`sage.plot.complex_plot.add_lightness_smoothing_to_rgb`

    ALGORITHM:

    Each pixel and lightness-delta is mapped from `(r, g, b, delta) \mapsto
    (h, l, s, delta)` using the standard RGB-to-HLS formula.

    Then the lightness is adjusted via `l \mapsto l' = l + 0.5 \cdot delta`.

    Finally map `(h, l', s) \mapsto (r, g, b)` using the standard HLS-to-RGB
    formula.

    EXAMPLES::

        sage: # needs numpy
        sage: import numpy as np
        sage: from sage.plot.complex_plot import add_contours_to_rgb
        sage: add_contours_to_rgb(np.array([[[0, 0.25, 0.5]]]),  # abs tol 1e-4
        ....:                     np.array([[0.75]]))
        array([[[0.25 , 0.625, 1.   ]]])
        sage: add_contours_to_rgb(np.array([[[0, 0, 0]]]),  # abs tol 1e-4
        ....:                     np.array([[1]]))
        array([[[0.5, 0.5, 0.5]]])
        sage: add_contours_to_rgb(np.array([[[1, 1, 1]]]),  # abs tol 1e-4
        ....:                     np.array([[-0.5]]))
        array([[[0.75, 0.75, 0.75]]])

    Raising ``dark_rate`` leads to bigger adjustments::

        sage: add_contours_to_rgb(np.array([[[0.5, 0.5, 0.5]]]),  # abs tol 1e-4        # needs numpy
        ....:                     np.array([[0.5]]), dark_rate=0.1)
        array([[[0.55, 0.55, 0.55]]])
        sage: add_contours_to_rgb(np.array([[[0.5, 0.5, 0.5]]]),  # abs tol 1e-4        # needs numpy
        ....:                     np.array([[0.5]]), dark_rate=0.5)
        array([[[0.75, 0.75, 0.75]]])
    """
    import numpy as np
    hls = rgb_to_hls(rgb)
    hls[..., 1] += dark_rate * delta
    hls = np.clip(hls, 0.0, 1.0)
    rgb = hls_to_rgb(hls)
    return rgb


class ComplexPlot(GraphicPrimitive):
    """
    The GraphicsPrimitive to display complex functions in using the domain
    coloring method

    INPUT:

    - ``rgb_data`` -- an array of colored points to be plotted

    - ``x_range`` -- a minimum and maximum x value for the plot

    - ``y_range`` -- a minimum and maximum y value for the plot

    TESTS::

        sage: p = complex_plot(lambda z: z^2-1, (-2, 2), (-2, 2))
    """
    def __init__(self, rgb_data, x_range, y_range, options):
        """
        TESTS::

            sage: p = complex_plot(lambda z: z^2-1, (-2, 2), (-2, 2))
        """
        self.x_range = x_range
        self.y_range = y_range
        self.x_count = len(rgb_data)
        self.y_count = len(rgb_data[0])
        self.rgb_data = rgb_data
        GraphicPrimitive.__init__(self, options)

    def get_minmax_data(self):
        """
        Return a dictionary with the bounding box data.

        EXAMPLES::

            sage: p = complex_plot(lambda z: z, (-1, 2), (-3, 4))
            sage: sorted(p.get_minmax_data().items())
            [('xmax', 2.0), ('xmin', -1.0), ('ymax', 4.0), ('ymin', -3.0)]
            sage: p = complex_plot(lambda z: z, (1, 2), (3, 4))
            sage: sorted(p.get_minmax_data().items())
            [('xmax', 2.0), ('xmin', 1.0), ('ymax', 4.0), ('ymin', 3.0)]
        """
        from sage.plot.plot import minmax_data
        return minmax_data(self.x_range, self.y_range, dict=True)

    def _allowed_options(self):
        """
        TESTS::

            sage: isinstance(complex_plot(lambda z: z, (-1,1), (-1,1))[0]._allowed_options(), dict)
            True
        """
        return {'plot_points':'How many points to use for plotting precision',
                'interpolation':'What interpolation method to use'}

    def _repr_(self):
        """
        TESTS::

            sage: isinstance(complex_plot(lambda z: z, (-1,1), (-1,1))[0]._repr_(), str)
            True
        """
        return "ComplexPlot defined by a %s x %s data grid" % (self.x_count, self.y_count)

    def _render_on_subplot(self, subplot):
        """
        TESTS::

            sage: complex_plot(lambda x: x^2, (-5, 5), (-5, 5))
            Graphics object consisting of 1 graphics primitive
        """
        options = self.options()
        x0, x1 = float(self.x_range[0]), float(self.x_range[1])
        y0, y1 = float(self.y_range[0]), float(self.y_range[1])
        subplot.imshow(self.rgb_data, origin='lower', extent=(x0, x1, y0, y1),
                       interpolation=options['interpolation'])


@options(plot_points=100, interpolation='catrom')
def complex_plot(f, x_range, y_range, contoured=False, tiled=False, cmap=None,
                 contour_type='logarithmic', contour_base=None, dark_rate=0.5,
                 nphases=10, **options):
    r"""
    ``complex_plot`` takes a complex function of one variable,
    `f(z)` and plots output of the function over the specified
    ``x_range`` and ``y_range`` as demonstrated below. The magnitude of
    the output is indicated by the brightness and the argument is
    represented by the hue.

    By default, zero magnitude corresponds to black output, infinite
    magnitude corresponds to white output. The options ``contoured``,
    ``tiled``, and ``cmap`` affect the output.

    ``complex_plot(f, (xmin, xmax), (ymin, ymax), contoured, tiled, cmap, ...)``

    INPUT:

    - ``f`` -- a function of a single complex value `x + iy`

    - ``(xmin, xmax)`` -- 2-tuple, the range of ``x`` values

    - ``(ymin, ymax)`` -- 2-tuple, the range of ``y`` values

    - ``cmap`` -- ``None``, or the string name of a matplotlib colormap, or an
      instance of a matplotlib Colormap, or the special string ``'matplotlib'``
      (default: ``None``); if ``None``, then hues are chosen from a standard
      color wheel, cycling from red to yellow to blue. If ``matplotlib``, then
      hues are chosen from a preset matplotlib colormap.

    The following named parameter inputs can be used to add contours and adjust
    their distribution:

    - ``contoured`` -- boolean (default: ``False``); causes the magnitude
      to be indicated by logarithmically spaced 'contours'. The
      magnitude along one contour is either twice or half the magnitude
      along adjacent contours.

    - ``dark_rate`` -- a positive number (default: 0.5); affects how quickly
      magnitudes affect how light/dark the image is. When there are contours,
      this affects how visible each contour is. Large values (near 1.0) have
      very strong, immediate effects, while small values (near 0.0) have
      gradual effects.

    - ``tiled`` -- boolean (default: ``False``); causes the magnitude to
      be indicated by logarithmically spaced 'contours' as in
      ``contoured``, and in addition for there to be `10` evenly
      spaced phase contours.

    - ``nphases`` -- positive integer (default: 10); when ``tiled=True``,
      this is the number of divisions the phase is divided into

    - ``contour_type`` -- either ``'logarithmic'``, or ``'linear'`` (default:
      ``'logarithmic'``); causes added contours to be of given type when
      ``contoured=True``.

    - ``contour_base`` -- positive integer; when ``contour_type`` is
      ``'logarithmic'``, this sets logarithmic contours at multiples of
      ``contour_base`` apart. When ``contour_type`` is ``'linear'``, this sets
      contours at distances of ``contour_base`` apart. If ``None``, then a
      default is chosen depending on ``contour_type``.

    The following inputs may also be passed in as named parameters:

    - ``plot_points`` -- integer (default: 100); number of points to
      plot in each direction of the grid

    - ``interpolation`` -- string (default: ``'catrom'``); the interpolation
      method to use: ``'bilinear'``, ``'bicubic'``, ``'spline16'``,
      ``'spline36'``, ``'quadric'``, ``'gaussian'``, ``'sinc'``,
      ``'bessel'``, ``'mitchell'``, ``'lanczos'``, ``'catrom'``,
      ``'hermite'``, ``'hanning'``, ``'hamming'``, ``'kaiser'``

    Any additional parameters will be passed to ``show()``, as long as they're
    valid.

    .. NOTE::

        Matplotlib colormaps can be chosen or customized to cater to different
        types of vision. The colormaps 'cividis' and 'viridis' in matplotlib
        are designed to be perceptually uniform to a broader audience. The
        colormap 'turbo' is similar to the default but with more even contrast.
        See [NAR2018]_ for more information about colormap choice for
        scientific visualization.

    EXAMPLES:

    Here we plot a couple of simple functions::

        sage: complex_plot(sqrt(x), (-5, 5), (-5, 5))                                   # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(sqrt(x), (-5, 5), (-5, 5)))

    ::

        sage: complex_plot(sin(x), (-5, 5), (-5, 5))                                    # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(sin(x), (-5, 5), (-5, 5)))

    ::

        sage: complex_plot(log(x), (-10, 10), (-10, 10))                                # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(log(x), (-10, 10), (-10, 10)))

    ::

        sage: complex_plot(exp(x), (-10, 10), (-10, 10))                                # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(exp(x), (-10, 10), (-10, 10)))

    A plot with a different choice of colormap::

        sage: complex_plot(exp(x), (-10, 10), (-10, 10), cmap='viridis')                # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(exp(x), (-10, 10), (-10, 10), cmap='viridis'))

    A function with some nice zeros and a pole::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (-3, 3), (-3, 3))                                         # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (-3, 3), (-3, 3)))

    The same function as above, but with contours. Contours render poorly with
    few plot points, so we use 300 here::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (-3, 3), (-3, 3), plot_points=300, contoured=True)        # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (-3, 3), (-3, 3), plot_points=300, contoured=True))

    The same function as above, but tiled and with the *plasma* colormap::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (-3, 3), (-3, 3),                                         # needs sage.symbolic
        ....:              plot_points=300, tiled=True, cmap='plasma')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (-3, 3), (-3, 3), plot_points=300, tiled=True, cmap='plasma'))

    When using ``tiled=True``, the number of phase subdivisions can be
    controlled by adjusting ``nphases``. We make the same plot with fewer
    tilings::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (-3, 3), (-3, 3), plot_points=300,                        # needs sage.symbolic
        ....:              tiled=True, nphases=5, cmap='plasma')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (-3, 3), (-3, 3), plot_points=300, tiled=True, nphases=5, cmap='plasma'))

    It is also possible to use *linear* contours. We plot the same function
    above on an inset, setting contours to appear `1` apart::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (0, 1), (0, 1), plot_points=300,                          # needs sage.symbolic
        ....:              contoured=True, contour_type='linear', contour_base=1)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (0, 1), (0, 1), plot_points=300, contoured=True, contour_type='linear', contour_base=1))

    Note that tightly spaced contours can lead to Moiré patterns and aliasing
    problems. For example::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (-3, 3), (-3, 3), plot_points=300,                        # needs sage.symbolic
        ....:              contoured=True, contour_type='linear', contour_base=1)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (-3, 3), (-3, 3), plot_points=300, contoured=True, contour_type='linear', contour_base=1))

    When choosing colormaps, cyclic colormaps such as *twilight* or *hsv* might
    be considered more appropriate for showing changes in phase without sharp
    color contrasts::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (-3, 3), (-3, 3), plot_points=300, cmap='twilight')       # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (-3, 3), (-3, 3), plot_points=300, cmap='twilight'))

    Passing *matplotlib* as the colormap gives a special colormap that is
    similar to the default::

        sage: f(z) = z^5 + z - 1 + 1/z                                                  # needs sage.symbolic
        sage: complex_plot(f, (-3, 3), (-3, 3),                                         # needs sage.symbolic
        ....:              plot_points=300, contoured=True, cmap='matplotlib')
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        def f(z): return z**5 + z - 1 + 1/z
        sphinx_plot(complex_plot(f, (-3, 3), (-3, 3), plot_points=300, contoured=True, cmap='matplotlib'))

    Here is the identity, useful for seeing what values map to what colors::

        sage: complex_plot(lambda z: z, (-3, 3), (-3, 3))                               # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(lambda z: z, (-3, 3), (-3, 3)))

    The Riemann Zeta function::

        sage: complex_plot(zeta, (-30,30), (-30,30))                                    # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(zeta, (-30,30), (-30,30)))

    For advanced usage, it is possible to tweak many parameters. Increasing
    ``dark_rate`` will make regions become darker/lighter faster when there are no
    contours::

        sage: complex_plot(zeta, (-30, 30), (-30, 30), dark_rate=1.0)                   # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(zeta, (-30,30), (-30,30), dark_rate=1.0))

    Decreasing ``dark_rate`` has the opposite effect. When there are contours,
    adjust ``dark_rate`` affects how visible contours are. Compare::

        sage: complex_plot(zeta, (-1, 9), (10, 20), plot_points=200,    # long time, needs sage.symbolic
        ....:              contoured=True, cmap='twilight', dark_rate=0.2)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(zeta, (-1, 9), (10, 20), plot_points=200, contoured=True, cmap='twilight', dark_rate=0.2))

    and::

        sage: complex_plot(zeta, (-1, 9), (10, 20), plot_points=200,    # long time, needs sage.symbolic
        ....:              contoured=True, cmap='twilight', dark_rate=0.75)
        Graphics object consisting of 1 graphics primitive

    .. PLOT::

        sphinx_plot(complex_plot(zeta, (-1, 9), (10, 20), plot_points=200, contoured=True, cmap='twilight', dark_rate=0.75))

    In practice, different values of ``dark_rate`` will work well with
    different colormaps.

    Extra options will get passed on to show(), as long as they are valid::

        sage: complex_plot(lambda z: z, (-3, 3), (-3, 3), figsize=[1,1])                # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive

    ::

        sage: complex_plot(lambda z: z, (-3, 3), (-3, 3)).show(figsize=[1,1])  # These are equivalent                   # needs sage.symbolic

    REFERENCES:

    Plotting complex functions with colormaps follows the strategy
    from [LD2021]_ and incorporates contour techniques described
    in [WegSem2010]_.

    TESTS:

    Test to make sure that using fast_callable functions works::

        sage: # needs sage.symbolic
        sage: f(x) = x^2
        sage: g = fast_callable(f, domain=CC, vars='x')
        sage: h = fast_callable(f, domain=CDF, vars='x')
        sage: P = complex_plot(f, (-10, 10), (-10, 10))
        sage: Q = complex_plot(g, (-10, 10), (-10, 10))
        sage: R = complex_plot(h, (-10, 10), (-10, 10))
        sage: S = complex_plot(exp(x)-sin(x), (-10, 10), (-10, 10))
        sage: P; Q; R; S
        Graphics object consisting of 1 graphics primitive
        Graphics object consisting of 1 graphics primitive
        Graphics object consisting of 1 graphics primitive
        Graphics object consisting of 1 graphics primitive

    Test to make sure symbolic functions still work without declaring
    a variable.  (We don't do this in practice because it doesn't use
    fast_callable, so it is much slower.)

    ::

        sage: complex_plot(sqrt, (-5, 5), (-5, 5))                                      # needs sage.symbolic
        Graphics object consisting of 1 graphics primitive
    """
    import matplotlib as mpl
    import numpy as np
    from sage.plot.graphics import Graphics
    from sage.plot.misc import setup_for_eval_on_grid
    from sage.ext.fast_callable import fast_callable
    from sage.rings.complex_double import CDF

    try:
        f = fast_callable(f, domain=CDF, expect_one_var=True)
    except (AttributeError, TypeError, ValueError):
        pass

    cdef double x, y
    _, ranges = setup_for_eval_on_grid([], [x_range, y_range],
                                       options['plot_points'])
    x_range = ranges[0]
    y_range = ranges[1]
    cdef list z_values = []
    cdef list row
    for y in srange(*y_range, include_endpoint=True):
        row = []
        for x in srange(*x_range, include_endpoint=True):
            sig_check()
            row.append(f(new_CDF_element(x, y)))
        z_values.append(row)

    if cmap is None:
        # produce colors using the established default method
        rgbs = complex_to_rgb(
            z_values, contoured=contoured, tiled=tiled,
            contour_type=contour_type, contour_base=contour_base,
            dark_rate=dark_rate, nphases=nphases
        )
    else:
        # choose colors from colormap
        if isinstance(cmap, str):
            if cmap == 'matplotlib':
                domain = np.linspace(0, 1, 256)
                shifted_domain = np.roll(domain, 128)
                default_cmap = mpl.colors.LinearSegmentedColormap.from_list(
                    "sage_default", mpl.colormaps['hsv'](shifted_domain)
                )
                cmap = default_cmap
            else:
                cmap = mpl.colormaps[cmap]
        rgbs = complex_to_cmap_rgb(
            z_values, cmap=cmap, contoured=contoured, tiled=tiled,
            contour_type=contour_type, contour_base=contour_base,
            dark_rate=dark_rate, nphases=nphases
        )

    g = Graphics()
    g._set_extra_kwds(Graphics._extract_kwds_for_show(options, ignore=['xmin', 'xmax']))
    g.add_primitive(ComplexPlot(rgbs, x_range[:2], y_range[:2], options))
    return g


def rgb_to_hls(rgb):
    r"""
    Convert array of rgb values (each in the range `[0, 1]`)
    to a numpy array of hls values (each in the range `[0, 1]`)

    INPUT:

    - ``rgb`` -- an `N \times 3` array of floats with values
      in the range `[0, 1]`; the rgb values at each point. (Note that the input
      can actually be of any dimension, such as `N \times M \times 3`, as long
      as the last dimension has length `3`).

    OUTPUT:

    An `N \times 3` Numpy array of floats in the range `[0, 1]`, with
    the same dimensions as the input array.

    .. SEEALSO::

        :func:`sage.plot.complex_plot.hls_to_rgb`

    EXAMPLES:

    We convert a row of floats and verify that we can convert back using
    ``hls_to_rgb``::

        sage: from sage.plot.complex_plot import rgb_to_hls, hls_to_rgb
        sage: rgb = [[0.2, 0.4, 0.5], [0.1, 0.3, 1.0]]
        sage: hls = rgb_to_hls(rgb)
        sage: hls   # abs tol 1e-4
        array([[0.55555556, 0.35      , 0.42857143],
               [0.62962963, 0.55      , 1.        ]])
        sage: hls_to_rgb(hls)  # abs tol 1e-4
        array([[0.2, 0.4, 0.5],
               [0.1, 0.3, 1. ]])

    Multidimensional inputs can be given as well::

        sage: multidim_arr = [[[0, 0.2, 0.4], [1, 1, 1]], [[0, 0, 0], [0.5, 0.6, 0.9]]]
        sage: rgb_to_hls(multidim_arr)   # abs tol 1e-4
        array([[[0.58333333, 0.2       , 1.        ],
                [0.        , 1.        , 0.        ]],
               [[0.        , 0.        , 0.        ],
                [0.625     , 0.7       , 0.66666667]]])
    """
    # Notation and algorithm corresponds to colorsys.rgb_to_hls
    import numpy as np
    rgb = np.asarray(rgb)
    if rgb.shape[-1] != 3:
        raise ValueError("Last dimension of input array must be 3; "
                         "shape {} was found.".format(rgb.shape))
    in_shape = rgb.shape
    rgb = np.asarray(rgb, dtype=np.dtype(float))
    rgb_max = rgb.max(-1)
    rgb_min = rgb.min(-1)
    l = (rgb_max + rgb_min)/2.0  # lightness

    hls = np.zeros_like(rgb)
    delta = np.ptp(rgb, -1)
    s = np.zeros_like(delta)

    ipos = delta > 0
    idx = (l <= 0.5) & ipos
    s[idx] = delta[idx] / (rgb_max[idx] + rgb_min[idx])

    idx = (l > 0.5) & ipos
    s[idx] = delta[idx] / (2.0 - rgb_max[idx] - rgb_min[idx])  # saturation

    # red is max
    idx = (rgb[..., 0] == rgb_max) & ipos
    hls[idx, 0] = (rgb[idx, 1] - rgb[idx, 2]) / delta[idx]

    # green is max
    idx = (rgb[..., 1] == rgb_max) & ipos
    hls[idx, 0] = 2.0 + (rgb[idx, 2] - rgb[idx, 0]) / delta[idx]

    # blue is max
    idx = (rgb[..., 2] == rgb_max) & ipos
    hls[idx, 0] = 4.0 + (rgb[idx, 0] - rgb[idx, 1]) / delta[idx]

    hls[..., 0] = (hls[..., 0] / 6.0) % 1.0
    hls[..., 1] = l
    hls[..., 2] = s

    return hls.reshape(in_shape)


def hls_to_rgb(hls):
    r"""
    Convert array of hls values (each in the range `[0, 1]`)
    to a numpy array of rgb values (each in the range `[0, 1]`)

    INPUT:

    - ``hls`` -- an `N \times 3` array of floats in the range `[0, 1]`; the hls
      values at each point. (Note that the input can actually be of any
      dimension, such as `N \times M \times 3`, as long as the last dimension
      has length `3`).

    OUTPUT:

    An `N \times 3`  Numpy array of floats in the range `[0, 1]`, with
    the same dimensions as the input array.

    .. SEEALSO::

        :func:`sage.plot.complex_plot.rgb_to_hls`

    EXAMPLES:

    We convert a row of floats and verify that we can convert back using
    ``rgb_to_hls``::

        sage: from sage.plot.complex_plot import rgb_to_hls, hls_to_rgb
        sage: hls = [[0.2, 0.4, 0.5], [0.1, 0.3, 1.0]]
        sage: rgb = hls_to_rgb(hls)
        sage: rgb  # abs tol 1e-4
        array([[0.52, 0.6 , 0.2 ],
               [0.6 , 0.36, 0.  ]])
        sage: rgb_to_hls(rgb)  # abs tol 1e-4
        array([[0.2, 0.4, 0.5],
               [0.1, 0.3, 1. ]])

    Multidimensional inputs can be given as well::

        sage: multidim_arr = [[[0, 0.2, 0.4], [0, 1, 0]], [[0, 0, 0], [0.5, 0.6, 0.9]]]
        sage: hls_to_rgb(multidim_arr)  # abs tol 1e-4
        array([[[0.28, 0.12, 0.12],
                [1.  , 1.  , 1.  ]],
               [[0.  , 0.  , 0.  ],
                [0.24, 0.96, 0.96]]])
    """
    import numpy as np
    hls = np.asarray(hls)
    if hls.shape[-1] != 3:
        raise ValueError("Last dimension of input array must be 3; "
                         "shape {} was found.".format(hls.shape))
    in_shape = hls.shape
    hls = np.array(
        hls, copy=False, dtype=np.dtype(float), ndmin=2
    )
    rgb = np.zeros_like(hls)

    szero_idx = hls[..., 2] == 0
    rgb[szero_idx, 0] = hls[szero_idx, 1]
    rgb[szero_idx, 1] = hls[szero_idx, 1]
    rgb[szero_idx, 2] = hls[szero_idx, 1]

    snonzero_idx = hls[..., 2] > 0
    m1 = np.zeros_like(snonzero_idx, dtype=float)
    m2 = np.zeros_like(snonzero_idx, dtype=float)

    idx = (hls[..., 1] <= 0.5)
    m2[idx] = hls[idx, 1] * (1.0 + hls[idx, 2])

    idx = (hls[..., 1] > 0.5)
    m2[idx] = hls[idx, 1] + hls[idx, 2] - hls[idx, 1] * hls[idx, 2]

    m1 = 2 * hls[..., 1] - m2

    rgb[snonzero_idx, 0] = _v(m1, m2, hls[..., 0] + 1.0 / 3.0)[snonzero_idx]
    rgb[snonzero_idx, 1] = _v(m1, m2, hls[..., 0])[snonzero_idx]
    rgb[snonzero_idx, 2] = _v(m1, m2, hls[..., 0] - 1.0 / 3.0)[snonzero_idx]

    return rgb.reshape(in_shape)


def _v(m1, m2, hue):
    """
    A helper function to convert from hls to rgb.

    This is a numpy version of colorsys._v.

    INPUT:

    - ``m1`` -- an array of floats with values in the range `[0, 1]`

    - ``m2`` -- an array of floats with values in the range `[0, 1]` with
      the same dimensions as ``m1``

    - ``hue`` -- an array of floats with values in the range `[0, 1]` with
      the same dimensions as ``m1``

    OUTPUT:

    A Numpy array of floats with values in the range `[0, 1]`. The dimensions
    are the same as the input arrays.

    EXAMPLES::

        sage: from sage.plot.complex_plot import _v
        sage: _v([0.1, 0.2], [0.1, 0.3], [0.25, 0.75])  # abs tol 1e-4
        array([0.1, 0.2])
    """
    import numpy as np
    m1 = np.asarray(m1, dtype=float)
    m2 = np.asarray(m2, dtype=float)
    hue = np.asarray(hue, dtype=float)

    if not m1.shape == m2.shape and m1.shape == hue.shape:
        raise ValueError(
            "Incompatible shapes given to _v. "
            "shapes {}, {}, {} given.".format(m1.shape, m2.shape, hue.shape)
        )

    out = np.zeros_like(m1, dtype=float)
    hue = hue % 1.0

    conditions = [hue < 1.0 / 6.0, hue < 1.0 / 2.0, hue < 2.0 / 3.0]
    out = np.select(conditions, [
        m1 + (m2 - m1) * hue * 6.0,
        m2,
        m1 + (m2 - m1) * (2.0 / 3.0 - hue) * 6.0
    ], default=m1)
    return out



# distutils: libraries = flint
# distutils: depends = flint/fmpz_mod_vec.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void _fmpz_mod_vec_set_fmpz_vec(fmpz * A, const fmpz * B, slong len, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_neg(fmpz * A, const fmpz * B, slong len, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_add(fmpz * a, const fmpz * b, const fmpz * c, slong n, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_sub(fmpz * a, const fmpz * b, const fmpz * c, slong n, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_scalar_mul_fmpz_mod(fmpz * A, const fmpz * B, slong len, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_scalar_addmul_fmpz_mod(fmpz * A, const fmpz * B, slong len, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_scalar_div_fmpz_mod(fmpz * A, const fmpz * B, slong len, const fmpz_t c, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_dot(fmpz_t d, const fmpz * A, const fmpz * B, slong len, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_dot_rev(fmpz_t d, const fmpz * A, const fmpz * B, slong len, const fmpz_mod_ctx_t ctx) noexcept
    void _fmpz_mod_vec_mul(fmpz * A, const fmpz * B, const fmpz * C, slong len, const fmpz_mod_ctx_t ctx) noexcept
